"""This is a catchall module for helper functions."""
from __future__ import absolute_import

from . import bootstrap
from . import misc
from . import stats
from . import video
from . import dataload
from . import whisk_permissions

# shortcuts
from .misc import rint, pick, pick_rows, printnow, globjoin


"""Module containing bootstrap methods for estimating differences between
groups. Loosely based on Efron 1983.
"""
from __future__ import print_function
from __future__ import division
from builtins import range
from builtins import object
from past.utils import old_div

import numpy as np
import pandas

def bootstrap_rms_distance(full_distribution, subset, n_boots=1000, seed=0):
    """Test whether subset of points comes from the larger set
    
    full_distribution : array of shape (n_points, dimensionality)
        The full set of vectors
    subset : array of shape (len_subset, dimensionality)
        A subset of the vectors
    
    The null hypothesis is that the subset comes from the full distribution.
    We take the mean Euclidean distance between each point in the subset 
    and the center of the full distribution as the test statistic.
    
    We then draw `n_boots` fake subsets with replacement from the full
    distribution. The same test statistic is calculated for each fake
    subset. A p-value is obtained by the fraction of draws that are more
    extreme than the true data.
    
    Returns: p_value, subset_distances, bootstrapped_distance_distribution
    """
    np.random.seed(seed)
    
    # true mean of the distribution
    distribution_mean = np.mean(full_distribution, axis=0)
    
    # Draw from the full distribution
    # Each draw contains the same number of samples as the dataset
    # There are `n_boots` draws total (one per row)
    idxs_by_boot = np.random.randint(0, len(full_distribution), 
        (n_boots, len(subset)))
    
    # Actual drawing
    # This will have shape (n_boots, len_dataset, dimensionality)
    draws_by_boot = np.array([
        full_distribution[row_idxs] for row_idxs in idxs_by_boot])
    
    # RMS distance of each row (dim2) from the average
    # This will have shape (n_boots, len_dataset)
    distances_by_boot = np.sqrt(np.mean(
        (draws_by_boot - [[distribution_mean]])**2, axis=2))
    true_distances = np.sqrt(np.mean(
        (subset - [distribution_mean])**2, axis=1))
    
    # Mean RMS distance of each boot (shape n_boots)
    mdistances_by_boot = np.mean(distances_by_boot, axis=1)
    
    # Mean RMS distance of the true subset
    true_mdistance = np.mean(true_distances)
    
    # Now we just take the z-score of the mean distance of the real dataset
    abs_deviations = np.abs(mdistances_by_boot - mdistances_by_boot.mean())
    true_abs_deviation = np.abs(true_mdistance - mdistances_by_boot.mean())
    p_value = old_div(np.sum(true_abs_deviation <= abs_deviations), float(n_boots))
    
    return p_value, true_distances, mdistances_by_boot
    

def pvalue_of_distribution(data, compare=0, floor=True, verbose=True):
    """Returns the two-tailed p-value of `compare` in `data`.
    
    First we choose the more extreme direction: the minimum of 
    (proportion of data points less than compare, 
    proportion of data points greater than compare).
    Then we double this proportion to make it two-tailed.
    
    floor : if True and the p-value is 0, use 2 / len(data)
        This is to account for the fact that outcomes of probability
        less than 1/len(data) will probably not occur in the sample.
    verbose : if the p-value is floored, print a warning
    
    Not totally sure about this, first of all there is some selection
    bias by choosing the more extreme comparison. Secondly, this seems to
    be the pvalue of obtaining 0 from `data`, but what we really want is the
    pvalue of obtaining `data` if the true value is zero.
    
    Probably better to obtain a p-value from permutation test or some other
    test on the underlying data.
    """
    n_more_extreme = np.sum(data < compare)
    cdf_at_value = old_div(n_more_extreme, float(len(data)))
    p_at_value = 2 * np.min([cdf_at_value, 1 - cdf_at_value])    
    
    # Optionally deal with p = 0
    if floor and (n_more_extreme == 0 or n_more_extreme == len(data)):
        p_at_value = old_div(2, float(len(data)))
        
        if verbose:
            print("warning: exactly zero p-value encountered in " + \
                "pvalue_of_distribution, flooring")
    
    return p_at_value

class DiffBootstrapper(object):
    """Object to estimate the difference between two groups with bootstrapping."""
    def __init__(self, data1, data2, n_boots=1000, min_bucket=5):
        self.data1 = data1
        self.data2 = data2
        self.n_boots = n_boots
        self.min_bucket = min_bucket
    
    def execute(self, seed=0):
        """Test the difference in means with bootstrapping.
        
        Data is drawn randomly from group1 and group2, with resampling.
        From these bootstraps, estimates with confidence intervals are 
        calculated for the mean of each group and the difference in means.
        
        The estimated difference is positive if group2 > group1.
        
        Sets: mean1, CI_1, mean2, CI_2, diff_estimate, diff_CI, p1, p2
        
        p1 is the p-value estimated from the distribution of differences
        p2 is the p-value from a 1-sample ttest on that distribution
        """
        if len(self.data1) < self.min_bucket or len(self.data2) < self.min_bucket:
            #~ raise BootstrapError(
                #~ 'insufficient data in bucket in bootstrap_two_groups')
            raise ValueError(
                'insufficient data in bucket in bootstrap_two_groups')
        
        if seed is not None:
            np.random.seed(seed)
        
        # Generate random samples, shape (n_boots, len(group))
        self.idxs1 = np.random.randint(0, len(self.data1), 
            (self.n_boots, len(self.data1)))
        self.idxs2 = np.random.randint(0, len(self.data2), 
            (self.n_boots, len(self.data2)))
        
        # Draw from the data
        self.draws1 = self.data1[self.idxs1]
        self.draws2 = self.data2[self.idxs2]
        
        # Bootstrapped means of each group
        self.means1 = self.draws1.mean(axis=1)
        self.means2 = self.draws2.mean(axis=1)
        
        # CIs on group means
        self.CI_1 = np.percentile(self.means1, (2.5, 97.5))
        self.CI_2 = np.percentile(self.means2, (2.5, 97.5))
        
        # Bootstrapped difference between the groups
        self.diffs = self.means2 - self.means1
        self.CI_diff = np.percentile(self.diffs, (2.5, 97.5))
        
        # p-value
        self.p_from_dist = pvalue_of_distribution(self.diffs, 0)
        
        # save memory
        del self.idxs1
        del self.idxs2
        del self.draws1
        del self.draws2
    
    @property
    def summary_group1(self):
        """Return mean, CI_low, CI_high on group1"""
        return self.means1.mean(), self.CI_1[0], self.CI_1[1]
    
    @property
    def summary_group2(self):
        return self.means2.mean(), self.CI_2[0], self.CI_2[1]
    
    @property
    def summary_diff(self):
        return self.diffs.mean(), self.CI_diff[0], self.CI_diff[1]
    
    @property
    def summary(self):
        return list(self.summary_group1) + list(self.summary_group2) + \
            list(self.summary_diff) + [self.p_from_dist]



# Stimulus-equalizing bootstrap functions

def difference_CI_bootstrap_wrapper(data, **boot_kwargs):
    """Given parsed data from single ulabel, return difference CIs.
    
    data : same format as bootstrap_main_effect expects
    
    Will calculate the following statistics:
        means : mean of each condition, across draws
        CIs : confidence intervals on each condition
        mean_difference : mean difference between conditions
        difference_CI : confidence interval on difference between conditions
        p : two-tailed p-value of 'no difference'
    
    Returns:
        dict of those statistics
    """
    # Yields a 1000 x 2 x N_trials matrix:
    # 1000 draws from the original data, under both conditions.
    bh = bootstrap_main_effect(data, meth=keep, **boot_kwargs)

    # Find the distribution of means of each draw, across trials
    # This is 1000 x 2, one for each condition
    # hist(means_of_all_draws) shows the comparison across conditions
    means_of_all_draws = bh.mean(axis=2)

    # Confidence intervals across the draw means for each condition
    condition_CIs = np.array([
        np.percentile(dist, (2.5, 97.5)) for dist in means_of_all_draws.T])

    # Means of each ulabel (centers of the CIs, basically)
    condition_means = means_of_all_draws.mean(axis=0)

    # Now the CI on the *difference between conditions*
    difference_of_conditions = np.diff(means_of_all_draws).flatten()
    difference_CI = np.percentile(difference_of_conditions, (2.5, 97.5)) 

    # p-value of 0. in the difference distribution
    cdf_at_value = old_div(np.sum(difference_of_conditions < 0.), \
        float(len(difference_of_conditions)))
    p_at_value = 2 * np.min([cdf_at_value, 1 - cdf_at_value])
    
    # Should probably floor the p-value at 1/n_boots

    return {'p' : p_at_value, 
        'means' : condition_means, 'CIs': condition_CIs,
        'mean_difference': difference_of_conditions.mean(), 
        'difference_CI' : difference_CI}

def bootstrap_main_effect(data, n_boots=1000, meth=None, min_bucket=5):
    """Given 2xN set of data of unequal sample sizes, bootstrap main effect.

    We will generate a bunch of fake datasets by resampling from data.
    Then we combine across categories. The total number of data points
    will be the same as in the original dataset; however, the resampling
    is such that each category is equally represented.    
    
    data : list of length N, each entry a list of length 2
        Each entry in `data` is a "category".
        Each category consists of two groups.
        The question is: what is the difference between the groups, without
        contaminating by the different size of each category?
    
    n_boots : number of times to randomly draw, should be as high as you
        can stand
    
    meth : what to apply to the drawn samples from each group
        If None, use means_tester
        It can be any function that takes (group0, group1)
        Results of every call are returned

    Returns:
        np.asarray([meth(group0, group1) for group0, group1 in each boot])    
    """    
    if meth is None:
        meth = means_tester
    
    # Convert to standard format
    data = [[np.asarray(d) for d in dd] for dd in data]
    
    # Test
    alld = np.concatenate([np.concatenate([dd for dd in d]) for d in data])
    if len(np.unique(alld)) == 0:
        raise BootstrapError("no data")
    elif len(np.unique(alld)) == 1:
        raise BootstrapError("all data points are identical")
    
    # How many to generate from each group, total
    N_group0 = np.sum([len(category[0]) for category in data])
    N_group1 = np.sum([len(category[1]) for category in data])
    N_categories = len(data)
    
    # Which categories to draw from
    res_l = []
    for n_boot in range(n_boots):
        # Determine the representation of each category
        # Randomly generating so in the limit each category is equally
        # represented. Alternatively we could use fixed, equal representation,
        # but then we need to worry about rounding error when it doesn't
        # divide exactly evenly.
        fakedata_category_label_group0 = np.random.randint(  
            0, N_categories, N_group0)
        fakedata_category_label_group1 = np.random.randint(
            0, N_categories, N_group1)
        
        # Draw the data, separately by each category
        fakedata_by_group = [[], []]
        for category_num in range(N_categories):
            # Group 0
            n_draw = np.sum(fakedata_category_label_group0 == category_num)
            if len(data[category_num][0]) < min_bucket:
                raise BootstrapError("insufficient data in a category")
            idxs = np.random.randint(0, len(data[category_num][0]),
                n_draw)
            fakedata_by_group[0].append(data[category_num][0][idxs])
            
            # Group 1
            n_draw = np.sum(fakedata_category_label_group0 == category_num)
            if len(data[category_num][1]) < min_bucket:
                raise BootstrapError("insufficient data in a category")
            idxs = np.random.randint(0, len(data[category_num][1]),
                n_draw)
            fakedata_by_group[1].append(data[category_num][1][idxs])
        
        # Concatenate across categories
        fakedata_by_group[0] = np.concatenate(fakedata_by_group[0])
        fakedata_by_group[1] = np.concatenate(fakedata_by_group[1])
        
        # Test difference in means
        #res = np.mean(fakedata_by_group[1]) - np.mean(fakedata_by_group[0])
        res = meth(fakedata_by_group[0], fakedata_by_group[1])
        res_l.append(res)
    
    return np.asarray(res_l)

# Lambdas for bootstrap_main_effect
def means_tester(d0, d1):
    return np.mean(d1) - np.mean(d0)

def keep(d0, d1):
    return (d0, d1)


# Utility bootstrap functions
def CI_compare(CI1, CI2):
    """Return +1 if CI1 > CI2, -1 if CI1 < CI2, 0 if overlapping"""
    if CI1[1] < CI2[0]:
        return -1
    elif CI2[1] < CI1[0]:
        return +1
    else:
        return 0

def simple_bootstrap(data, n_boots=1000, min_bucket=20):
    if len(data) < min_bucket:
        raise BootstrapError("too few samples")
    
    res = []
    data = np.asarray(data)
    for boot in range(n_boots):
        idxs = np.random.randint(0, len(data), len(data))
        draw = data[idxs]
        res.append(np.mean(draw))
    res = np.asarray(res)
    CI = np.percentile(res, (2.5, 97.5))
    
    return res, res.mean(), CI

class BootstrapError(BaseException):
    pass

def bootstrap_CIs_on_dataframe(df):
    """Returns CIs on the mean of each row
    
    Here's how to do it with a groupby:
        res = gobj.apply(lambda df: wwutils.bootstrap.bootstrap_CIs_on_dataframe(df.T))
    
    Returns: DataFrame
        Index will be the same as df.index
        Columns will be ['lo', 'mean', 'hi', 'mpl_lo', 'mpl_hi']
        matplotlib-style error bars can be extracted as:
        np.array([res['mpl_lo'].values, res['mpl_hi'].values])
    """
    # CIs
    CI_l = []
    CI_keys_l = []
    for idx, sub_data in df.iterrows():
        CI = simple_bootstrap(sub_data)[2]
        CI_l.append(CI)
        CI_keys_l.append(idx)

    #~ agg_err = pandas.DataFrame(CI_l, columns=['lo', 'hi'], 
        #~ index=pandas.Index(CI_keys_l, names=df.index.names))
    
    # The index should be the same as df.index
    agg_err = pandas.DataFrame(CI_l, columns=['lo', 'hi'], 
        index=df.index)
    
    # Mean
    agg_err['mean'] = df.mean(1)
    
    # matplotlib error bars
    agg_err['mpl_lo'] = agg_err['mean'] - agg_err['lo']
    agg_err['mpl_hi'] = agg_err['hi'] - agg_err['mean']
    return agg_err



import os
import sys
from ctypes import CDLL

def get_current_conda_env_name():
    """
    Returns the name of the current conda environment.
    """
    conda_prefix = os.getenv('CONDA_PREFIX')
    if conda_prefix:
        # The last directory in the CONDA_PREFIX path is the name of the current environment
        return os.path.basename(conda_prefix)
    return None

def load_ffmpeg_dlls_from_conda_env():
    # Get the Conda prefix for the environment
    conda_prefix = os.getenv('CONDA_PREFIX')
    
    # If CONDA_PREFIX is not set, try to derive it from sys.prefix
    if not conda_prefix:
        conda_prefix = sys.prefix
        
    # If still not found, return an error
    if not conda_prefix:
        raise EnvironmentError("Conda environment not found")
        
    # Construct the path to the Library\bin directory inside the Conda environment
    library_bin_path = os.path.join(conda_prefix, 'Library', 'bin')
    
    # List of FFmpeg DLLs to be loaded
    if os.name == 'posix':
      
      ffmpeg_dll_names = [
          "libavcodec.so.60",
          "libavdevice.so.60",
          "libavformat.so.60",
          "libavutil.so.58",
          "libswscale.so.7",
      ]      
      
    elif os.name == 'nt':
        
        ffmpeg_dll_names = [
            "avcodec-60.dll",
            "avdevice-60.dll",
            "avformat-60.dll",
            "avutil-58.dll",
            "swscale-7.dll"
        ]
    
    for dll_name in ffmpeg_dll_names:
        dll_path = os.path.join(library_bin_path, dll_name)
        if os.path.exists(dll_path):
            CDLL(dll_path)
            print(f"Loaded {dll_path} successfully!")
        else:
            raise FileNotFoundError(f"Could not find {dll_name} at {dll_path}")

# Example usage
current_env = get_current_conda_env_name()
if current_env:
    print(f"Current Conda environment: {current_env}")
    load_ffmpeg_dlls_from_conda_env()
else:
    print("Not running in a Conda environment")



"""Utility functions for WhiskiWrap

FileNamer defines the naming convention for whiski-related files. This is
not really mandatory but is used by the pipeline_trace function in base.

The other utility functions are for probing the availability of needed
system commands: ffmpeg, trace.
"""

import os
import subprocess

class FileNamer(object):
    """Defines the naming convention for whiski-related files.

    This can be initialized from a basename, such as:
        fn = FileNamer('~/my_directory/session_name')
    or from an existing video file or whiskers file, such as:
        fn = FileNamer.from_whiskers('~/my_directory/session_name.whiskers')
    In the latter case a warning is issued if no such file exists, or if it
    does not follow the typical naming convention.

    Once initialized, this object generates names:
        fn.whiskers
        fn.tiff_stack
        fn.video(type='mp4')
    """
    def __init__(self, basename):
        """Initialize based on full path and filename (without extension)."""
        self.basename = os.path.abspath(os.path.expanduser(basename))

    def video(self, typ='tif'):
        return self.basename + '.' + typ

    @property
    def tiff_stack(self):
        """Returns the name for the tiff stack"""
        return self.video('tif')

    @property
    def whiskers(self):
        """Return the name for the whiskers file"""
        return self.basename + '.whiskers'

    @property
    def measurements(self):
        """Return the name for the measurements file"""
        return self.basename + '.measurements'

    @classmethod
    def from_video(self, video_name):
        """Generates FileNamer based on an existing video name"""
        if not os.path.exists(video_name):
            print("warning: nonexistent video %s" % video_name)
        basename, ext = os.path.splitext(video_name)
        if ext not in ['.mp4', '.avi', '.mkv', '.tif']:
            print("warning: %s does not appear to be a video file" % video_name)
        return FileNamer(basename)

    @classmethod
    def from_whiskers(self, whiskers_file_name):
        """Generates FileNamer based on an existing whiskers file"""
        if not os.path.exists(whiskers_file_name):
            print("warning: nonexistent whiskers file %s" % whiskers_file_name)
        basename, ext = os.path.splitext(whiskers_file_name)
        if ext != '.whiskers':
            raise ValueError("%s is not a whiskers file" % whiskers_file_name)
        return FileNamer(basename)

    @classmethod
    def from_tiff_stack(self, tiff_stack_filename):
        """Generates FileNamer based on an existing tiff stack"""
        #~ if not os.path.exists(tiff_stack_filename):
            #~ print "warning: nonexistent tiff stack %s" % tiff_stack_filename
        basename, ext = os.path.splitext(tiff_stack_filename)
        if ext != '.tif':
            raise ValueError("%s is not a *.tif stack" % tiff_stack_filename)
        return FileNamer(basename)

    @property
    def hdf5(self):
        """Return the name for the hdf5 file"""
        return self.basename + '.hdf5'


def probe_command_availability(cmd):
    """Try to run 'cmd' in a subprocess and return availability.

    'cmd' should be provided in the format expected by subprocess: a string,
    or a list of strings if multiple arguments.

    Raises RuntimeError if the called process crashes (eg, via Ctrl+C)

    Returns:
        command_available, stdout, stderr

    stdout and stderr will be '' if command was not available.
    """
    # Try to initialize a pipe which will only work if it is available
    command_available = True
    try:
        # If it fails here due to nonexistence of command, pipe is
        # never initialized
        pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        return False, '', ''

    # This extracts data from the pipe
    try:
        stdout, stderr = pipe.communicate()
    except Exception as e:
        raise RuntimeError("process crashed") from e

    # Try to terminate it if it didn't already happen
    try:
        pipe.terminate()
    except OSError:
        pass

    return command_available, stdout, stderr

def probe_needed_commands(*commands, **kwargs):
    """Test whether we have the commands we need.

    By default checks 'ffmpeg' and 'trace'.

    Additional commands can be specified as arguments.

    Optional keyword arguments:
        paths (list): Specific paths to check for the commands.
    """
    default_commands = ['ffmpeg', 'trace']
    all_commands = default_commands + list(commands)
    paths = kwargs.get('paths', [])

    for cmd in all_commands:
        cmd_av = False, '', ''
        
        # Check specified paths first
        for path in paths:
            full_cmd = [os.path.join(path, cmd)]
            cmd_av = probe_command_availability(full_cmd)
            if cmd_av[0]:
                break
        
        # If not found in specified paths, check system path
        if not cmd_av[0]:
            cmd_av = probe_command_availability(cmd)

        if not cmd_av[0]:
            raise OSError(f"'{cmd}' is not available on the system path or specified paths")
        



import WhiskiWrap as ww
import sys, os, re, json
import numpy as np
import pandas as pd
# import time
from scipy.signal import argrelextrema
from scipy.stats import gaussian_kde

def get_summary(filename, filter = True):
    # Check if it's a list of whiskers/measurements files
    if isinstance(filename, list):
        if filename[0].endswith(('.whiskers', '.measurements')):
        # if file is a list of whiskers/measurements files, load all of them
            summary = ww.read_whiskers_measurements(filename)
            # update nomenclature
            if 'whisker_id' in summary:
                # replace whisker_id with wid, and frame_id with fid
                summary.rename(columns={'whisker_id': 'wid', 'frame_id': 'fid'}, inplace=True)
            filename = filename[0]
            base_filename = os.path.basename(filename.replace('_right', '').replace('_left', '')).split(".")[0]
            # also remove the trailing numbers at the end of the filename
            base_filename = re.sub(r'_\d+$', '', base_filename)
            # whiskerpad_file is in the directory above the input file
            whiskerpad_file = os.path.join(os.path.dirname(os.path.dirname(filename)), f'whiskerpad_{base_filename}.json')     

    elif isinstance(filename, str) and filename.endswith('.hdf5'):
    # if file is a hdf5 file, load it
        summary = ww.read_whiskers_hdf5_summary(filename)
        # print(summary.head())   

        base_filename = os.path.basename(filename.replace('_right', '').replace('_left', '')).split(".")[0]
        whiskerpad_file = os.path.join(os.path.dirname(filename), f'whiskerpad_{base_filename}.json')

    # Load whiskerpad json file
    with open(whiskerpad_file, 'r') as f:
        whiskerpad_params = json.load(f)

    # get the whiskerpad params for the correct side         
    if '_right' in str(filename):
        whiskerpad = next((whiskerpad for whiskerpad in whiskerpad_params['whiskerpads'] if whiskerpad['FaceSide'].lower() == 'right'), None)
    else:
        whiskerpad = next((whiskerpad for whiskerpad in whiskerpad_params['whiskerpads'] if whiskerpad['FaceSide'].lower() == 'left'), None)

    # get face axis and orientation 
    face_axis = whiskerpad['FaceAxis']
    face_orientation = whiskerpad['FaceOrientation']

    # Determine a threshold for whisker length
    length_threshold = determine_length_threshold(summary)
    # score_threshold = determine_score_threshold(summary)

    if filter:
        summary = filter_whiskers(summary, length_threshold)

    return summary

def reassess_whisker_ids(filename):

    summary = get_summary(filename)

    grouped_by_fid = group_summary_by_fid(summary)
    for fid in range(1, max(grouped_by_fid.keys())):
        current_frame_whiskers = grouped_by_fid[fid]
        # next_frame_whiskers = grouped_by_fid[fid + 1]
        previous_frame_whiskers = grouped_by_fid[fid - 1]
        
        # Change wids only if number and order varies between frames 
        if is_reassignment_needed(current_frame_whiskers, previous_frame_whiskers, face_axis, face_orientation):
            for whisker_index, whisker in current_frame_whiskers.iterrows():
                closest_whisker, distance =  find_closest_whisker(whisker, previous_frame_whiskers)

                #  Reassign only if wid is different and within limit
                if is_within_threshold(distance) and whisker['wid'] != closest_whisker['wid']:
                    reassign_wid(grouped_by_fid, fid, whisker_index, closest_whisker['wid'])
        else:
            continue

    return update_summary_with_new_ids(grouped_by_fid), filtered_summary

def is_reassignment_needed(current_frame_whiskers, previous_frame_whiskers, face_axis, face_orientation):
    # Check if the number of whiskers is the same in both frames
    # if len(current_frame_whiskers) != len(previous_frame_whiskers):
    #     if len(current_frame_whiskers) > len(previous_frame_whiskers):
    #         return True

    # Order check based on face axis and orientation
    if face_axis.lower() == 'vertical' and face_orientation.lower() == 'down':
        if not is_ascending_order(current_frame_whiskers, 'follicle_y'):
            return True
    elif face_axis.lower() == 'vertical' and face_orientation.lower() == 'up':
        if not is_descending_order(current_frame_whiskers, 'follicle_y'):
            return True
    elif face_axis.lower() == 'horizontal' and face_orientation.lower() == 'left':
        if not is_ascending_order(current_frame_whiskers, 'follicle_x'):
            return True
    elif face_axis.lower() == 'horizontal' and face_orientation.lower() == 'right':
        if not is_descending_order(current_frame_whiskers, 'follicle_x'):
            return True
    return False

def is_ascending_order(frame_whiskers, column):
    # Check if the values in the specified column are in ascending order
    y_values = frame_whiskers[column].tolist()
    return y_values == sorted(y_values)

def is_descending_order(frame_whiskers, column):
    # Check if the values in the specified column are in descending order
    y_values = frame_whiskers[column].tolist()
    return y_values == sorted(y_values, reverse=True)

def is_within_threshold(distance, threshold=10.0):
    return distance <= threshold

def reassign_wid(grouped_by_fid, current_fid, whisker_index, new_wid):
    current_frame = grouped_by_fid[current_fid]
    
    # Find the index of the whisker in the current frame that has the new_wid
    target_index = current_frame[current_frame['wid'] == new_wid].index

    # If a whisker with new_wid is found, swap the wids
    if not target_index.empty:
        # Get the current wid of the whisker at whisker_index
        current_wid = current_frame.loc[whisker_index, 'wid']
        # Assign it 
        grouped_by_fid[current_fid].loc[target_index[0], 'wid'] = current_wid

    # Then assign the new_wid to the whisker_index
    grouped_by_fid[current_fid].loc[whisker_index, 'wid'] = new_wid

def update_summary_with_new_ids(grouped_by_fid):
    updated_summary = pd.DataFrame()
    for fid, group in grouped_by_fid.items():
        updated_summary = pd.concat([updated_summary, group], ignore_index=True)
    return updated_summary

def find_closest_whisker(whisker, previous_frame_whiskers):
    # Method 1
    # time it
    # time1 = time.time()
    min_distance = float('inf')
    closest_whisker = None

    for _, next_whisker in previous_frame_whiskers.iterrows():
        distance = euclidean_distance(whisker, next_whisker)
        if distance < min_distance:
            min_distance = distance
            closest_whisker = next_whisker

    # print(f"Method 1 took {time.time() - time1} seconds")

    # time2 = time.time()
    # #  Method 2
    # closest_whisker, min_distance = closest_whisker_distance(whisker, previous_frame_whiskers)     
    # print(f"Method 2 took {time.time() - time2} seconds")

    return closest_whisker, min_distance

def closest_whisker_distance(whisker, previous_frame_whiskers):
    # Convert whisker to a NumPy array
    whisker_array = np.array([whisker['follicle_x'], whisker['follicle_y'], whisker['angle']])

    # Convert previous_frame_whiskers to a NumPy array
    previous_frame_whiskers_array = previous_frame_whiskers[['follicle_x', 'follicle_y', 'angle']].to_numpy()

    # Calculate the Euclidean distance to all whiskers in the previous frame
    distances = np.sqrt(np.sum((previous_frame_whiskers_array - whisker_array)**2, axis=1))

    # Find the index of the closest whisker
    min_index = np.argmin(distances)

    # Get the closest whisker and minimum distance
    closest_whisker = previous_frame_whiskers.iloc[min_index]
    min_distance = distances[min_index]

    return closest_whisker, min_distance

def euclidean_distance(whisker1, whisker2, method='follicle_angle'):
    if method == 'follicle_tip':
    # Using follicle and tip coordinates
        distance = np.sqrt((whisker1['follicle_x'] - whisker2['follicle_x'])**2 + 
                        (whisker1['follicle_y'] - whisker2['follicle_y'])**2 +
                        (whisker1['tip_x'] - whisker2['tip_x'])**2 +
                        (whisker1['tip_y'] - whisker2['tip_y'])**2)
    elif method == 'follicle_angle':
    # Using follicle coordinates and angle
        distance = np.sqrt((whisker1['follicle_x'] - whisker2['follicle_x'])**2 + 
                        (whisker1['follicle_y'] - whisker2['follicle_y'])**2 +
                        (angle_difference(whisker1['angle'], whisker2['angle']))**2)
    return distance

def angle_difference(angle1, angle2):
    # Calculate the minimum difference between two angles
    return min(abs(angle1 - angle2), 360 - abs(angle1 - angle2))

def extract_features(summary):
    # Extract and possibly create new features
    features = summary[['tip_x', 'tip_y', 'follicle_x', 'follicle_y']]
    return features

def cluster_whiskers(features):
    # Normalize features if necessary
    # features_normalized = normalize(features)

    # Apply clustering algorithm
    kmeans = KMeans(n_clusters=num_whiskers)  # num_whiskers is an estimated number of unique whiskers
    clusters = kmeans.fit_predict(features)

    return clusters

def reassess_whisker_ids_with_clustering(summary):
    features = extract_features(summary)
    clusters = cluster_whiskers(features)

    # Map clusters to new whisker IDs
    summary['new_wid'] = clusters

    return summary

def filter_whiskers(summary, length_threshold, score_threshold=None):
    if score_threshold is not None:
        return summary[(summary['pixel_length'] > length_threshold) & (summary['score'] > score_threshold)]
    else:
        return summary[summary['pixel_length'] > length_threshold]

def determine_length_threshold(summary):
    # Convert data to numpy array and reshape
    data = np.array(summary['pixel_length']).reshape(-1, 1).ravel()

    # Fit the KDE model
    kde = gaussian_kde(data)

    # Evaluate the densities
    kde_values = kde.evaluate(data)

    # Find local minima of the KDE, which are the troughs of the bimodal distribution
    local_minima = argrelextrema(kde_values, np.less)

    # If there are any local minima, use the first one as the cutoff
    if local_minima[0].size > 0:
        length_threshold = data[local_minima[0][0]]
    else:
        # If there are no local minima, fall back to using the 75th percentile
        length_threshold = np.percentile(summary['pixel_length'], 75)
    # #  Cutoff at 75th percentile, since most whiskers are actually hair
    # length_threshold = np.percentile(summary['pixel_length'], 75)

    # mean_length = np.mean(summary['pixel_length'])
    # std_dev_length = np.std(summary['pixel_length'])
    # length_threshold = mean_length - std_dev_length

    return length_threshold

def determine_score_threshold(summary):

    mean_score = np.mean(summary['score'])
    std_dev_score = np.std(summary['score'])
    score_threshold = mean_score + std_dev_score

    return score_threshold

def group_summary_by_fid(summary):
    """
    Groups whisker segments by frame ID (fid).

    Args:
        summary (DataFrame): The DataFrame containing whisker segment data.

    Returns:
        dict: A dictionary where each key is a frame ID and each value is a DataFrame of whisker segments for that frame.
    """
    # Group the DataFrame by 'fid' and create a dictionary
    # if 'fid' in summary.columns:
    grouped = summary.groupby('fid')
    # else:
    #     grouped = summary.groupby('frame_id')
    grouped_by_fid = {fid: group for fid, group in grouped}
    
    return grouped_by_fid

def plot_summary_distributions(summary, h5_filename):
    #  plot the distrbution of pixel_length, and score. 
    #  Save the plot to the same directory as the input file
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    ax[0].hist(summary['pixel_length'], bins=50)
    ax[0].set_title('Pixel Length Distribution')
    ax[0].set_xlabel('Pixel Length')
    ax[0].set_ylabel('Frequency')
    ax[1].hist(summary['score'], bins=50)
    ax[1].set_title('Score Distribution')
    ax[1].set_xlabel('Score')
    ax[1].set_ylabel('Frequency')
    plt.savefig(h5_filename.replace('.hdf5', '_summary.png'))
    plt.close()

def plot_angle_traces(loaded_data, output_filename):
    # plot wid 0, 1 and 2 angle traces
    import matplotlib.pyplot as plt
    # angle = loaded_data[loaded_data['wid'] == 0]['angle']
    # plt.plot(angle)
    # plt.title('Angle trace for wid 0')
    for wid in range(3):
        angle = loaded_data[loaded_data['wid'] == wid]['angle']
        plt.plot(angle)
    plt.title('Angle traces for wids 0, 1, and 2')
    plt.xlabel('Frame')
    plt.ylabel('Angle')
    plt.savefig(output_filename.replace('.hdf5', '_angle_traces.png'))
    plt.close()

def update_wids(h5_filename):
    # Call the reassess function
    updated_summary, filtered_summary = reassess_whisker_ids(h5_filename)

    # Save the updated summary to a new hdf5 file
    output_filename = h5_filename.replace('.hdf5', '_updated_wids.hdf5')

    # Create an HDF5 file and store the DataFrame
    updated_summary.to_hdf(output_filename, key='updated_summary', format='table', data_columns=True)

    # with pd.HDFStore(output_filename, 'w') as store:
    #     store.put('updated_summary', updated_summary, format='table', data_columns=True)

    print(f"Updated summary saved to {output_filename}")

    #  Load the updated summary and plot the distributions
    # with pd.HDFStore(output_filename, 'r') as store:
    #     loaded_data = store.get('updated_summary')
    plot_angle_traces(filtered_summary, h5_filename)
    plot_angle_traces(updated_summary, output_filename)

if __name__ == "__main__":
    #  get argument from command line
    h5_filename = sys.argv[1]
    update_wids(h5_filename)



"""Functions for running tests and/or benchmarks with WhiskiWrap.

Each test is run in a session directory containing all the necessary
files for that test.

The function run_standard_benchmarks goes through a suite of tests. This
tests that everything is working properly and also serves as an example
of how to choose parameters and trace input videos.
"""


import os
import numpy as np
import tables
import time
import pandas
import WhiskiWrap
import shutil


def setup_session_directory(directory, input_video, force=False):
    """Create (or overwrite) directory for whisker tracking"""
    # Parse the input video filename
    input_video = os.path.abspath(os.path.expanduser(input_video))
    if not os.path.exists(input_video):
        raise ValueError("%s does not exist" % input_video)
    input_video_directory, input_video_filename = os.path.split(input_video)
    
    # Erase existing directory and create anew
    whiski_files = ['.mp4', '.avi', '.whiskers', '.tif', '.measurements',
        '.detectorbank', '.parameters', '.hdf5']
    if os.path.exists(directory):
        # Check that it looks like a whiskers directory
        file_list = os.listdir(directory)
        for filename in file_list:
            if (os.path.splitext(filename)[1]) not in whiski_files:
                raise ValueError(directory + 
                    " does not look safe to overwrite, aborting")
        
        # Get user confirmation
        if not force:
            confirm = input('Ok to erase %s? [y/N]: ' % directory)
            if confirm.upper() != 'Y':
                raise ValueError("did not receive permission to setup test")
        
        # Erase
        os.system('rm -rf %s' % directory)
    os.mkdir(directory)
    
    # Copy the input video into the session directory
    new_video_filename = os.path.join(directory, input_video_filename)
    shutil.copyfile(input_video, new_video_filename)
    
    # Copy the parameter files in
    for filename in [WhiskiWrap.PARAMETERS_FILE, 
        WhiskiWrap.HALFSPACE_DB_FILE, WhiskiWrap.LINE_DB_FILE]:
        raw_filename = os.path.split(filename)[1]
        shutil.copyfile(filename, os.path.join(directory, raw_filename))
    
    return WhiskiWrap.utils.FileNamer.from_video(new_video_filename)

def run_benchmarks(benchmark_params, test_root, force=False):
    """Run the benchmarks
    
    For every row in benchmark params, run a trace on the input video
    using the params specified.
    
    benchmark_params: DataFrame with columns corresponding to keywords
        to pass to pipeline_trace. Should have columns 'name',
        'input_video', 'chunk_sz_frames', 'epoch_sz_frames',
        'frame_start', 'frame_stop', 'n_trace_processes', etc
    
    Returns:
        test_results, durations
        test_results : Dict from test['name'] to results read from hdf5 file
        durations : list of durations taken
    """
    WhiskiWrap.utils.probe_needed_commands()
    
    test_results = {}
    durations = []    
    for idx, test in benchmark_params.iterrows():
        print(test['name'])
        test_dir = os.path.expanduser(os.path.join(test_root, test['name']))
        fn = setup_session_directory(test_dir, test['input_video'], force=force)

        # Run
        start_time = time.time()
        WhiskiWrap.pipeline_trace(
            fn.video('mp4'),
            fn.hdf5,
            chunk_sz_frames=test['chunk_sz_frames'],
            epoch_sz_frames=test['epoch_sz_frames'],
            frame_start=test['frame_start'],
            frame_stop=test['frame_stop'],
            n_trace_processes=test['n_trace_processes'])
        stop_time = time.time()
        durations.append(stop_time - start_time)

        # Get the summary
        with tables.open_file(fn.hdf5) as fi:
            test_results[test['name']] = pandas.DataFrame.from_records(
                fi.root.summary.read()) 
    
    return test_results, durations

def run_standard_benchmarks(test_root='~/whiski_wrap_test', force=False,
    input_video=None, n_frames=None, epoch_sz=None,
    n_processes_l=(2, 4), chunk_size_l=(100, 300),
    ):
    """Run a suite of standard benchmarks.
    
    This function sets up a series of tests and then calls run_benchmarks
    on those tests.
    
    test_root : directory to store test results. Will be overwritten.
    force : if True, does not ask permission to overwrite anything
    input_video : which video to use as a test.
        Default: test_video165s.mp4 in the WhiskiWrap directory
    n_frames : number of frames to process
        Default: np.max(chunk_size_l) * np.max(n_process_l), that is, 
        just enough that the largest chunks will use all the processes
    epoch_sz : length of each epoch
        Default: equal to n_frames
    n_processes_l : a list of process numbers to test
    chunk_size_l : a list of chunk sizes to test
    
    The number of frames and the epoch length is chosen to be 
    np.max(chunk_size_l) * max_processes. It is assumed that additional
    epochs would take the same time.
    
    Returns: test_results, durations
    """
    # Check we have commands we need
    WhiskiWrap.utils.probe_needed_commands()
    
    # Set up test root
    test_root = normalize_path_and_optionally_get_permission(test_root,
        force=force)
    
    # Find the video to use
    if input_video is None:
        input_video = os.path.join(WhiskiWrap.DIRECTORY, 'test_video_165s.mp4')
    
    # Determine number of frames
    if n_frames is None:
        # Enough so that we use up all the processes on the largest chunk
        n_frames = np.max(n_processes_l) * np.max(chunk_size_l)
    if epoch_sz is None:
        epoch_sz = n_frames    
    
    # Construct the tests
    tests = []
    for chunk_sz in chunk_size_l:
        for n_process in n_processes_l:
            test_name = '%d_frames_%d_chunksz_%d_procs' % (
                n_frames, chunk_sz, n_process)
            tests.append([test_name, input_video, 0, n_frames, epoch_sz,
                chunk_sz, n_process])
    tests_df = pandas.DataFrame(tests, columns=(
        'name', 'input_video', 'frame_start', 'frame_stop', 
        'epoch_sz_frames', 'chunk_sz_frames', 'n_trace_processes'))

    # Run the tests
    test_results, durations = run_benchmarks(tests_df, test_root)
    tests_df['duration'] = durations

    return test_results, tests_df

def run_offset_test(test_root='~/whiski_wrap_test', start=1525, offset=5,
    n_frames=30, force=False):
    """Run a test where we offset the frame start"""
    # Check we have commands we need
    WhiskiWrap.utils.probe_needed_commands()
    
    # Set up test root
    test_root = normalize_path_and_optionally_get_permission(test_root,
        force=force)
    
    # Find the video to use
    vfile1 = os.path.join(WhiskiWrap.DIRECTORY, 'test_video2.mp4')

    # Construct the tests
    stop = start + n_frames
    tests = pandas.DataFrame([
        ['one_chunk', vfile1, start, stop, 100, 100, 1],
        ['one_chunk_offset', vfile1, start + offset, stop, 100, 100, 1],
        ],
        columns=(
            'name', 'input_video', 'frame_start', 'frame_stop', 
            'epoch_sz_frames', 'chunk_sz_frames', 'n_trace_processes'))
    
    # Run the tests
    test_results, durations = run_benchmarks(tests, test_root)

    return test_results, durations

def get_permission_for_test_root(test_root):
    """Ask for permission to run in test_root"""
    response = input('Run tests in %s? [y/N]: ' % test_root)
    if response.upper() != 'Y':
        raise ValueError("did not receive permission to run test")   

def normalize_path_and_optionally_get_permission(test_root, force=False):
    # Form test root
    test_root = os.path.abspath(os.path.expanduser(test_root))
    
    # Get permission to use it
    if not force:
        get_permission_for_test_root(test_root)    
    
    return test_root

def run_standard(test_root='~/whiski_wrap_test', force=False):
    """Run a standard trace on a test file to get baseline time"""
    # Check we have commands we need
    WhiskiWrap.utils.probe_needed_commands()
    
    # Set up test root
    test_root = normalize_path_and_optionally_get_permission(test_root,
        force=force)
    
    # Find the video to use
    vfile1 = os.path.join(WhiskiWrap.DIRECTORY, 'test_video_165s.mp4')    

    # Set up the test directory
    fn = setup_session_directory(os.path.join(test_root, 'standard'), vfile1)

    # Run the test
    start_time = time.time()
    WhiskiWrap.trace_chunk(fn.video('mp4'))
    stop_time = time.time()
    standard_duration = stop_time - start_time
    
    # Stitch
    WhiskiWrap.setup_hdf5(fn.hdf5, expectedrows=100000)
    WhiskiWrap.append_whiskers_to_hdf5(
        whisk_filename=fn.whiskers,
        h5_filename=fn.hdf5, 
        chunk_start=0)    
    
    # Get the result
    with tables.open_file(fn.hdf5) as fi:
        test_result = pandas.DataFrame.from_records(
            fi.root.summary.read())     

    return test_result, standard_duration


"""Catchall module within the catchall module for really one-off stuff."""
from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from builtins import map
from builtins import range
from past.builtins import basestring
from builtins import object
from past.utils import old_div

import numpy as np
import warnings
import matplotlib.mlab as mlab
import matplotlib # for spectrogrammer
import scipy.signal
import os
import re
import datetime
import glob
import wwutils
import pandas

## Deprecated stuff
from wwutils.video import OutOfFrames

def frame_dump(*args, **kwargs):
    warnings.warn("use wwutils.video instead of wwutils.misc", stacklevel=2)
    return wwutils.video.frame_dump(*args, **kwargs)

def frame_dump_pipe(*args, **kwargs):
    warnings.warn("use wwutils.video instead of wwutils.misc", stacklevel=2)
    return wwutils.video.get_frame(*args, **kwargs)
    
def process_chunks_of_video(*args, **kwargs):
    warnings.warn("use wwutils.video instead of wwutils.misc", stacklevel=2)
    return wwutils.video.process_chunks_of_video(*args, **kwargs)

def get_video_aspect(*args, **kwargs):
    warnings.warn("use wwutils.video instead of wwutils.misc", stacklevel=2)
    return wwutils.video.get_video_aspect(*args, **kwargs)

def get_video_duration(*args, **kwargs):
    warnings.warn("use wwutils.video instead of wwutils.misc", stacklevel=2)
    return wwutils.video.get_video_duration(*args, **kwargs)
##

def globjoin(dirname, pattern, normalize=True):
    """Join dirname to pattern, and glob it
    
    If normalize: calls os.path.abspath on every result
    """
    res = glob.glob(os.path.join(dirname, pattern))
    if normalize:
        res = list(map(os.path.abspath, res))
    return res

def time_of_file(filename, fmt='%Y%m%d%H%M%S'):
    """Return the modification time of the file as a datetime.
    
    If fmt is not None: apply strftime(fmt) and return the string
    """
    dt = datetime.datetime.fromtimestamp(os.path.getmtime(filename))
    
    if fmt is None:
        return dt
    else:
        return dt.strftime(fmt)


class Spectrogrammer(object):
    """Turns a waveform into a spectrogram"""
    def __init__(self, NFFT=256, downsample_ratio=1, new_bin_width_sec=None,
        max_freq=None, min_freq=None, Fs=1.0, noverlap=None, normalization=0,
        detrend=mlab.detrend_mean, **kwargs):
        """Object to turn waveforms into spectrograms.
        
        This is a wrapper around mlab.specgram. What this object provides
        is slightly more intelligent parameter choice, and a nicer way
        to trade off resolution in frequency and time. It also remembers
        parameter choices, so that the same object can be used to batch
        analyze a bunch of waveforms using the `transform` method.
        
        Arguments passed to mlab.specgram
        ----------------------------------
        NFFT - number of points used in each segment
            Determines the number of frequency bins, which will be
            NFFT / 2 before stripping out those outside min_freq and max_freq
        
        noverlap - int, number of samples of overlap between segments
            Default is NFFT / 2
        
        Fs - sampling rate
        
        detrend - detrend each segment before FFT
            Default is to remove the mean (DC component)
        
        **kwargs - anything else you want to pass to mlab.specgram
        
        
        Other arguments
        ---------------
        downsample_ratio - int, amount to downsample in time
            After all other calculations are done, the temporal resolution
        
        new_bin_width_sec - float, target temporal resolution
            The returned spectrogram will have a temporal resolution as
            close to this as possible.
            If this is specified, then the downsample_ratio is adjusted
            as necessary to achieve it. If noverlap is left as default,
            it will try 50% first and then 0, to achieve the desired resolution.
            If it is not possible to achieve within a factor of 2 of this
            resolution, a warning is issued.
        
        normalization - the power in each frequency bin is multiplied by
            the frequency raised to this power.
            0 means do nothing.
            1 means that 1/f noise becomes white.
        
        min_freq, max_freq - discard frequencies outside of this range
        
        
        Returns
        -------
        Pxx - 2d array of power in dB. Shape (n_freq_bins, n_time_bins)
            May contain -np.inf where the power was exactly zero.
        
        freqs - 1d array of frequency bins
        
        t - 1d array of times
        
        
        Theory
        ------
        The fundamental tradeoff is between time and frequency resolution and
        is set by NFFT.
        
        For instance, consider a 2-second signal, sampled at 1024Hz, chosen
        such that the number of samples is 2048 = 2**11.
        *   If NFFT is 2048, you will have 1024 frequency bins (spaced 
            between 0KHz and 0.512KHz) and 1 time bin. 
            This is a simple windowed FFT**2, with the redundant negative
            frequencies discarded since the waveform is real.
            Note that the phase information is lost.
        *   If NFFT is 256, you will have 128 frequency bins and 8 time bins.
        *   If NFFT is 16, you will have 8 freqency bins and 128 time bins.
        
        In each case the FFT-induced trade-off is:
            n_freq_bins * n_time_bins_per_s = Fs / 2
            n_freq_bins = NFFT / 2
        
        So far, using only NFFT, we have traded off time resolution for
        frequency resolution. We can achieve greater noise reduction with
        appropriate choice of noverlap and downsample_ratio. The PSD
        function achieves this by using overlapping segments, then averaging
        the FFT of each segment. The use of noverlap in mlab.specgram is 
        a bit of a misnomer, since no temporal averaging occurs there!
        But this object can reinstate this temporal averaging.
        
        For our signal above, if our desired temporal resolution is 64Hz,
        that is, 128 samples total, and NFFT is 16, we have a choice.
        *   noverlap = 0. Non-overlapping segments. As above, 8 frequency
            bins and 128 time bins. No averaging
        *   noverlap = 64. 50% overlap. Now we will get 256 time bins.
            We can then average together each pair of adjacent bins
            by downsampling, theoretically reducing the noise. Note that
            this will be a biased estimate since the overlapping windows
            are not redundant.
        *   noverlap = 127. Maximal overlap. Now we will get about 2048 bins,
            which we can then downsample by 128 times to get our desired
            time resolution.
        
        The trade-off is now:
            overlap_factor = (NFFT - overlap) / NFFT
            n_freq_bins * n_time_bins_per_s * overlap_factor = Fs / downsample_ratio / 2
        
        Since we always do the smoothing in the time domain, n_freq bins = NFFT / 2
        and the tradeoff becomes
            n_time_bins_per_s = Fs / downsample_ratio / (NFFT - overlap)
        
        That is, to increase the time resolution, we can:
            * Decrease the frequency resolution (NFFT)
            * Increase the overlap, up to a maximum of NFFT - 1
              This is a sort of spurious improvement because adjacent windows
              are highly correlated.
            * Decrease the downsample_ratio (less averaging)
        
        To decrease noise, we can:
            * Decrease the frequency resolution (NFFT)
            * Increase the downsample_ratio (more averaging, fewer timepoints)
        
        How to choose the overlap, or the downsample ratio? In general,
        50% overlap seems good, since we'd like to use some averaging, but
        we get limited benefit from averaging many redundant samples.        
        
        This object tries for 50% overlap and adjusts the downsample_ratio
        (averaging) to achieve the requested temporal resolution. If this is
        not possible, then no temporal averaging is done (just like mlab.specgram)
        and the overlap is increased as necessary to achieve the requested
        temporal resolution.
        """
        self.downsample_ratio = downsample_ratio # until set otherwise
        
        # figure out downsample_ratio
        if new_bin_width_sec is not None:
            # Set noverlap to default
            if noverlap is None:
                # Try to do it with 50% overlap
                noverlap = old_div(NFFT, 2)
            
            # Calculate downsample_ratio to achieve this
            self.downsample_ratio = \
                Fs * new_bin_width_sec / float(NFFT - noverlap)
            
            # If this is not achievable, then try again with minimal downsampling
            if np.rint(self.downsample_ratio).astype(np.int) < 1:
                self.downsample_ratio = 1
                noverlap = np.rint(NFFT - Fs * new_bin_width_sec).astype(np.int)
            
        # Convert to nearest int and test if possible
        self.downsample_ratio = np.rint(self.downsample_ratio).astype(np.int)        
        if self.downsample_ratio == 0:
            print("requested temporal resolution too high, using maximum")
            self.downsample_ratio = 1
    
        # Default value for noverlap if still None
        if noverlap is None:
            noverlap = old_div(NFFT, 2)
        self.noverlap = noverlap
        
        # store other defaults
        self.NFFT = NFFT
        self.max_freq = max_freq
        self.min_freq = min_freq
        self.Fs = Fs
        self.normalization = normalization
        self.detrend = detrend
        self.specgram_kwargs = kwargs

    
    def transform(self, waveform):
        """Converts a waveform to a suitable spectrogram.
        
        Removes high and low frequencies, rebins in time (via median)
        to reduce data size. Returned times are the midpoints of the new bins.
        
        Returns:  Pxx, freqs, t    
        Pxx is an array of dB power of the shape (len(freqs), len(t)).
        It will be real but may contain -infs due to log10
        """
        # For now use NFFT of 256 to get appropriately wide freq bands, then
        # downsample in time
        Pxx, freqs, t = mlab.specgram(waveform, NFFT=self.NFFT, 
            noverlap=self.noverlap, Fs=self.Fs, detrend=self.detrend, 
            **self.specgram_kwargs)
        
        # Apply the normalization
        Pxx = Pxx * np.tile(freqs[:, np.newaxis] ** self.normalization, 
            (1, Pxx.shape[1]))

        # strip out unused frequencies
        if self.max_freq is not None:
            Pxx = Pxx[freqs < self.max_freq, :]
            freqs = freqs[freqs < self.max_freq]
        if self.min_freq is not None:
            Pxx = Pxx[freqs > self.min_freq, :]
            freqs = freqs[freqs > self.min_freq]

        # Rebin in size "downsample_ratio". If last bin is not full, discard.
        Pxx_rebinned = []
        t_rebinned = []
        for n in range(0, len(t) - self.downsample_ratio + 1, 
            self.downsample_ratio):
            Pxx_rebinned.append(
                np.median(Pxx[:, n:n+self.downsample_ratio], axis=1).flatten())
            t_rebinned.append(
                np.mean(t[n:n+self.downsample_ratio]))

        # Convert to arrays
        Pxx_rebinned_a = np.transpose(np.array(Pxx_rebinned))
        t_rebinned_a = np.array(t_rebinned)

        # log it and deal with infs
        Pxx_rebinned_a_log = -np.inf * np.ones_like(Pxx_rebinned_a)
        Pxx_rebinned_a_log[np.nonzero(Pxx_rebinned_a)] = \
            10 * np.log10(Pxx_rebinned_a[np.nonzero(Pxx_rebinned_a)])


        self.freqs = freqs
        self.t = t_rebinned_a
        return Pxx_rebinned_a_log, freqs, t_rebinned_a

def fix_pandas_display_width(meth=1, dw=0, max_columns=12):
    """Various fixes"""
    import pandas
    if meth == 0:
        pandas.set_option('display.width', 0)
    elif meth == 1:
        pandas.set_option('display.max_columns', max_columns)
        pandas.set_option('display.width', None)


class UniquenessError(Exception):
    pass

def only_one(l):
    """Returns the only value in l, or l itself if non-iterable.
    
    Compare 'unique_or_error', which allows multiple identical etnries.
    """
    # listify
    if not hasattr(l, '__len__'):
        l = [l]
    
    # check length
    if len(l) != 1:
        raise UniquenessError("must contain exactly one value; instead: %r" % l)
    
    # return entry
    return l[0]

def unique_or_error(a):
    """Asserts that `a` contains only one unique value and returns it
    
    Compare 'only_one' which does not allow repeats.
    """    
    u = np.unique(np.asarray(a))
    if len(u) == 0:
        raise UniquenessError("no unique values found, should be one")
    if len(u) > 1:
        raise UniquenessError("%d unique values found, should be one" % len(u))
    else:
        return u[0]

def printnow(s):
    """Write string to stdout and flush immediately"""
    import sys
    sys.stdout.write(str(s) + "\n")
    sys.stdout.flush()

def get_file_time(filename, human=False):
    import time
    # Get modification time
    res = os.path.getmtime(filename)
    
    # Convert to human-readable
    if human:
        res = time.ctime(res)
    return res

def pickle_load(filename):
    import pickle
    with open(filename, 'rb') as fi:
        res = pickle.load(fi)
    return res

def pickle_dump(obj, filename):
    import pickle
    with open(filename, 'wb') as fi:
        pickle.dump(obj, fi)

def invert_linear_poly(p):
    """Helper function for inverting fit.coeffs"""
    return old_div(np.array([1, -p[1]]).astype(float), p[0])

def apply_and_filter_by_regex(pattern, list_of_strings, sort=True):
    """Apply regex pattern to each string and return result.
    
    Non-matches are ignored.
    If multiple matches, the first is returned.
    """
    res = []
    for s in list_of_strings:
        m = re.match(pattern, s)
        if m is None:
            continue
        else:
            res.append(m.groups()[0])
    if sort:
        return sorted(res)
    else:
        return res

def regex_filter(pattern, list_of_strings):
    """Apply regex pattern to each string and return those that match.
    
    See also regex_capture
    """
    return [s for s in list_of_strings if re.match(pattern, s) is not None]

def regex_capture(pattern, list_of_strings, take_index=0):
    """Apply regex pattern to each string and return a captured group.

    Same as old apply_and_filter_by_regex, but without the sorting.
    See also regex_filter. This will match that order.
    """
    # Apply filter to each string
    res_l = []
    for s in list_of_strings:
        m = re.match(pattern, s)
        
        # Append the capture, if any
        if m is not None:
            res_l.append(m.groups()[take_index])
    
    return res_l
    

def rint(arr):
    """Round with rint and cast to int
    
    If `arr` contains NaN, casting it to int causes a spuriously negative
    number, because NaN cannot be an int. In this case we raise ValueError.
    """
    if np.any(np.isnan(np.asarray(arr))):
        raise ValueError("cannot convert arrays containing NaN to int")
    return np.rint(arr).astype(np.int)

def is_nonstring_iter(val):
    """Check if the input is iterable, but not a string.
    
    Recently changed this to work for Unicode. 
    This should catch a subset of the old way, because previously Unicode
    strings caused this to return True, but now they should return False.
    
    Will print a warning if this is not the case.
    """
    # Old way
    res1 = hasattr(val, '__len__') and not isinstance(val, str)
    
    # New way
    res2 = hasattr(val, '__len__') and not isinstance(val, basestring)
    
    if res2 and not res1:
        print("warning: check is_nonstring_iter")
    
    return res2

def pick(df, isnotnull=None, **kwargs):
    """Function to pick row indices from DataFrame.
    
    Copied from kkpandas
    
    This method provides a nicer interface to choose rows from a DataFrame
    that satisfy specified constraints on the columns.
    
    isnotnull : column name, or list of column names, that should not be null.
        See pandas.isnull for a defintion of null
    
    All additional kwargs are interpreted as {column_name: acceptable_values}.
    For each column_name, acceptable_values in kwargs.items():
        The returned indices into column_name must contain one of the items
        in acceptable_values.
    
    If acceptable_values is None, then that test is skipped.
        Note that this means there is currently no way to select rows that
        ARE none in some column.
    
    If acceptable_values is a single string or value (instead of a list), 
    then the returned rows must contain that single string or value.
    
    TODO:
    add flags for string behavior, AND/OR behavior, error if item not found,
    return unique, ....
    """
    msk = np.ones(len(df), dtype=np.bool)
    for key, val in list(kwargs.items()):
        if val is None:
            continue
        elif is_nonstring_iter(val):
            msk &= df[key].isin(val)
        else:
            msk &= (df[key] == val)
    
    if isnotnull is not None:
        # Edge case
        if not is_nonstring_iter(isnotnull):
            isnotnull = [isnotnull]
        
        # Filter by not null
        for key in isnotnull:
            msk &= -pandas.isnull(df[key])

    return df.index[msk]

def pick_rows(df, **kwargs):
    """Returns sliced DataFrame based on indexes from pick"""
    return df.loc[pick(df, **kwargs)]

def no_warn_rs():
    warnings.filterwarnings('ignore', module='ns5_process.RecordingSession$')    

def parse_by_block(lb_counts, pb_counts, lb_trial_numbers, pb_trial_numbers,
    start_trial=None, last_trial=None, session_name=None):
    """Parses counts into each block (with corresponding trial numbers)
    
    Using the known block structure (80 trials LB, 80 trials PB, etc) 
    and the trial labels in the folded, split the counts into each 
    sequential block.
    
    parse_folded_by_block could almost be reimplemented with this function
    if only it accepted __getslice__[trial_number]
    Instead the code is almost duplicated.
    
    This one is overridden because it starts at 157:
        YT6A_120201_behaving, start at 161
    
    lb_counts, pb_counts : Array-like, list of counts, one per trial
    lb_trial_numbers, pb_trial_numbers : Array-like, same as counts, but
        contianing trial numbers.
    start_trial : where to start counting up by 80
        if None, auto-get from session_db and unit_db (or just use 1 if
        session name not provided)
    last_trial : last trial to include, inclusive
        if None, use the max trial in either set of labels
    session_name : if start_trial is None and you specify this, it will
        auto grab start_trial from session_db
    
    Returns: counts_by_block
    A list of arrays, always beginning with LB, eg LB1, PB1, LB2, PB2...
    """
    # Auto-get first trial
    if start_trial is None:
        if session_name is None:
            start_trial = 1
        elif session_name == 'YT6A_120201_behaving':
            # Forcible override
            start_trial = 161
        else:
            import my.dataload
            session_db = my.dataload.getstarted()['session_db']
            first_trial = int(round(session_db['first_trial'][session_name]))
            # Convert to beginning of first LBPB with any trials
            # The first block might be quite short
            # Change the final +1 to +161 to start at the first full block
            start_trial = (old_div((first_trial - 1), 160)) * 160 + 1
    
    # Arrayify
    lb_counts = np.asarray(lb_counts)
    pb_counts = np.asarray(pb_counts)
    lb_trial_numbers = np.asarray(lb_trial_numbers)
    pb_trial_numbers = np.asarray(pb_trial_numbers)
    
    # Where to stop putting trials into blocks    
    if last_trial is None:
        last_trial = np.max([lb_trial_numbers.max(), pb_trial_numbers.max()])
    
    # Initialize return variable
    res_by_block = []
    
    # Parse by block
    for block_start in range(start_trial, last_trial + 1, 80):
        # Counts from lb in this block
        lb_this_block_msk = (
            (lb_trial_numbers >= block_start) &
            (lb_trial_numbers < block_start + 80))
        lb_this_block = lb_counts[lb_this_block_msk]
        
        # Counts from pb in this block
        pb_this_block_msk = (
            (pb_trial_numbers >= block_start) &
            (pb_trial_numbers < block_start + 80))
        pb_this_block = pb_counts[pb_this_block_msk]
        
        # Error check
        if np.mod(block_start - start_trial, 160) == 0:
            # Should be in an LB block
            assert len(pb_this_block) == 0
            if len(lb_this_block) == 0:
                print("warning: no trials around trial %d" % block_start)
            res_by_block.append(lb_this_block)
        else:
            # Should be in a PB block
            assert len(lb_this_block) == 0
            if len(pb_this_block) == 0:
                print("warning: no trials around trial %d" % block_start)            
            res_by_block.append(pb_this_block)
    
    # Error check that all counts were included and ordering maintained
    assert np.all(np.concatenate(res_by_block[::2]) == 
        lb_counts[lb_trial_numbers >= start_trial])
    assert np.all(np.concatenate(res_by_block[1::2]) == 
        pb_counts[pb_trial_numbers >= start_trial])
    
    return res_by_block 

def parse_folded_by_block(lb_folded, pb_folded, start_trial=1, last_trial=None,
    session_name=None):
    """Parses Folded into each block
    
    parse_by_block is now more feature-ful
    TODO: reimplement parse_by_block to just return trial numbers, then
    this function can wrap that and use the trial numbers to slice the foldeds
    
    Using the known block structure (80 trials LB, 80 trials PB, etc) 
    and the trial labels in the folded, split the counts into each 
    sequential block.
    
    lb_folded, pb_folded : Folded with `label` attribute set with trial number
    start_trial : where to start counting up by 80
    last_trial : last trial to include, inclusive
        if None, use the max trial in either set of labels
    session_name : no longer used. should make this load trials_info and
        grab first trial, if anything.
    
    Returns: counts_by_block
    A list of arrays, always beginning with LB, eg LB1, PB1, LB2, PB2...
    """
    # Override start trial for some munged sessions
    #~ if session_name in ['YT6A_120201_behaving', 'CR24A_121019_001_behaving']:
        #~ print "overriding start trial"
        #~ start_trial = 161
    
    # Where to stop putting trials into blocks    
    if last_trial is None:
        last_trial = np.max([lb_folded.labels.max(), pb_folded.labels.max()])
    
    # Initialize return variable
    res_by_block = []
    
    # Parse by block
    for block_start in range(start_trial, last_trial + 1, 80):
        # Counts from lb in this block
        lb_this_block_msk = (
            (lb_folded.labels >= block_start) &
            (lb_folded.labels < block_start + 80))
        lb_this_block = lb_folded.get_slice(lb_this_block_msk)
        
        # Counts from pb in this block
        pb_this_block_msk = (
            (pb_folded.labels >= block_start) &
            (pb_folded.labels < block_start + 80))
        pb_this_block = pb_folded.get_slice(pb_this_block_msk)
        
        # Error check
        if np.mod(block_start - start_trial, 160) == 0:
            # Should be in an LB block
            assert len(pb_this_block) == 0
            #assert len(lb_this_block) > 0
            if len(lb_this_block) == 0:
                print("warning: no trials around trial %d" % block_start)
            res_by_block.append(lb_this_block)
        else:
            # Should be in a PB block
            assert len(lb_this_block) == 0
            #assert len(pb_this_block) > 0
            if len(pb_this_block) == 0:
                print("warning: no trials around trial %d" % block_start)            
            res_by_block.append(pb_this_block)
    
    return res_by_block 


def yoked_zscore(list_of_arrays, axis=1):
    """Concatenate arrays, z-score together, break apart again"""
    concatted = np.concatenate(list_of_arrays, axis=axis)
    means = np.mean(concatted, axis=axis)
    stdevs = np.std(concatted, axis=axis)
    
    res = []
    for arr in list_of_arrays:
        if axis == 1:
            res.append(old_div((arr - means[:, None]), stdevs[:, None]))
        elif axis == 0:
            res.append(old_div((arr - means[None, :]), stdevs[None, :]))
        else:
            raise ValueError("axis must be 0 or 1")
    return res


def gaussian_smooth(signal, gstd=100, glen=None, axis=1, **filtfilt_kwargs):
    """Smooth a signal with a Gaussian window
    
    signal : array-like, to be filtered
    gstd : standard deviation of Gaussian in samples (can be float)
    glen : half-width of (truncated) Gaussian
        Default is int(2.5 * gstd)
        If you are using padding (on by default) and the pad length which
        is a function of `glen` is longer than the data, you will get a
        smoothing error. Lower `glen` or lower `padlen`.
    axis : 0 or 1
        Default is to filter the columns of 2d data
    filtfilt_kwargs : other kwargs to pass to filtfilt
        padtype - 'odd', 'even', 'constant', None
            Default is 'odd', that is, continuing the signal at either end
            with odd symmetry
        padlen - int or None
            Default is None, which is 3 * max(len(signal), glen)
    
    NaNs will cause problems. You should probably interpolate them, using
    perhaps interp_nans in this module.
    """
    import scipy.signal
    
    # Defaults
    signal = np.asarray(signal)
    if glen is None:    
        glen = int(2.5 * gstd)
    
    # Incantation such that b[0] == 1.0
    b = scipy.signal.gaussian(glen * 2, gstd, sym=False)[glen:]
    b = old_div(b, b.sum())
    
    # Smooth
    if signal.ndim == 1:
        res = scipy.signal.filtfilt(b, [1], signal, **filtfilt_kwargs)
    elif signal.ndim == 2:
        if axis == 0:
            res = np.array([scipy.signal.filtfilt(b, [1], sig, **filtfilt_kwargs) 
                for sig in signal])
        elif axis == 1:
            res = np.array([scipy.signal.filtfilt(b, [1], sig, **filtfilt_kwargs) 
                for sig in signal.T]).T
        else:
            raise ValueError("axis must be 0 or 1")
    else:
        raise ValueError("signal must be 1d or 2d")
    
    return res

def interp_nans(signal, axis=1, left=None, right=None, dtype=float):
    """Replaces nans in signal by interpolation along axis
    
    signal : array-like, containing NaNs
    axis : 0 or 1
        Default is to interpolate along columns
    left, right : to be passed to interp
    dtype : Signal is first converted to this type, mainly to avoid
        conversion to np.object
    """
    # Convert to array
    res = np.asarray(signal, dtype=float).copy()

    # 1d or 2d behavior
    if res.ndim == 1:
        # Inner loop
        nan_mask = np.isnan(res)
        res[nan_mask] = np.interp(
            np.where(nan_mask)[0], # x-coordinates where we need a y
            np.where(~nan_mask)[0], # x-coordinates where we know y
            res[~nan_mask], # known y-coordinates
            left=left, right=right)
    elif res.ndim == 2:
        if axis == 0:
            res = np.array([
                interp_nans(sig, left=left, right=right, dtype=dtype)
                for sig in res])
        elif axis == 1:
            res = np.array([
                interp_nans(sig, left=left, right=right, dtype=dtype)
                for sig in res.T]).T
        else:
            raise ValueError("axis must be 0 or 1")
    else:
        raise ValueError("signal must be 1d or 2d")
    return res


# Correlation and coherence functions
def correlate(v0, v1, mode='valid', normalize=True, auto=False):
    """Wrapper around np.correlate to calculate the timepoints

    'full' : all possible overlaps, from last of first and beginning of
        second, to vice versa. Total length: 2*N - 1
    'same' : Slice out the central 'N' of 'full'. There will be one more
        negative than positive timepoint.
    'valid' : only overlaps where all of both arrays are included
    
    normalize: accounts for the amount of data in each bin
    auto: sets the center peak to zero
    
    Positive peaks (latter half of the array) mean that the second array
    leads the first array.
   
    """
    counts = np.correlate(v0, v1, mode=mode)
    
    if len(v0) != len(v1):
        raise ValueError('not tested')
    
    if mode == 'full':
        corrn = np.arange(-len(v0) + 1, len(v0), dtype=np.int)
    elif mode == 'same':
        corrn = np.arange(old_div(-len(v0), 2), len(v0) - (old_div(len(v0), 2)), 
            dtype=np.int)
    else:
        raise ValueError('mode not tested')
    
    if normalize:
        counts = old_div(counts, (len(v0) - np.abs(corrn)).astype(float))
    
    if auto:
        counts[corrn == 0] = 0
    
    return counts, corrn

def binned_pair2cxy(binned0, binned1, Fs=1000., NFFT=256, noverlap=None,
    windw=mlab.window_hanning, detrend=mlab.detrend_mean, freq_high=100,
    average_over_trials=True):
    """Given 2d array of binned times, return Cxy
    
    Helper function to ensure faking goes smoothly
    binned: 2d array, trials on rows, timepoints on cols
        Keep trials lined up!
    rest : psd_kwargs. noverlap defaults to NFFT/2

    Trial averaging, if any, is done between calculating the spectra and
    normalizing them.

    Will Cxy each trial with psd_kwargs, then mean over trials, then slice out 
    frequencies below freq_high and return.
    """
    # Set up psd_kwargs
    if noverlap is None:
        noverlap = old_div(NFFT, 2)
    psd_kwargs = {'Fs': Fs, 'NFFT': NFFT, 'noverlap': noverlap, 
        'detrend': detrend, 'window': windw}

    # Cxy each trial
    ppxx_l, ppyy_l, ppxy_l = [], [], []
    for row0, row1 in zip(binned0, binned1):
        ppxx, freqs = mlab.psd(row0, **psd_kwargs)
        ppyy, freqs = mlab.psd(row1, **psd_kwargs)
        ppxy, freqs = mlab.csd(row0, row1, **psd_kwargs)
        ppxx_l.append(ppxx); ppyy_l.append(ppyy); ppxy_l.append(ppxy)
    
    # Optionally mean over trials, then normalize
    S12 = np.real_if_close(np.array(ppxy_l))
    S1 = np.array(ppxx_l)
    S2 = np.array(ppyy_l)
    if average_over_trials:
        S12 = S12.mean(0)
        S1 = S1.mean(0)
        S2 = S2.mean(0)
    Cxy = old_div(S12, np.sqrt(S1 * S2))
    
    # Truncate unnecessary frequencies
    if freq_high:
        topbin = np.where(freqs > freq_high)[0][0]
        freqs = freqs.T[1:topbin].T
        Cxy = Cxy.T[1:topbin].T
    return Cxy, freqs

def binned2pxx(binned, Fs=1000., NFFT=256, noverlap=None,
    windw=mlab.window_hanning, detrend=mlab.detrend_mean, freq_high=100):
    """Given 2d array of binned times, return Pxx
    
    Helper function to ensure faking goes smoothly
    binned: 2d array, trials on rows, timepoints on cols
    rest : psd_kwargs. noverlap defaults to NFFT/2
    
    Will Pxx each trial separately with psd_kwargs, then slice out 
    frequencies below freq_high and return.
    """
    # Set up psd_kwargs
    if noverlap is None:
        noverlap = old_div(NFFT, 2)
    psd_kwargs = {'Fs': Fs, 'NFFT': NFFT, 'noverlap': noverlap, 
        'detrend': detrend, 'window': windw}    
    
    # Pxx each trial
    ppxx_l = []
    for row in binned:
        ppxx, freqs = mlab.psd(row, **psd_kwargs)
        ppxx_l.append(ppxx)
    
    # Truncate unnecessary frequencies
    if freq_high:
        topbin = np.where(freqs > freq_high)[0][0]
        freqs = freqs[1:topbin]
        Pxx = np.asarray(ppxx_l)[:, 1:topbin]       
    return Pxx, freqs


def sem(data, axis=None):
    """Standard error of the mean"""
    if axis is None:
        N = len(data)
    else:
        N = np.asarray(data).shape[axis]
    
    return old_div(np.std(np.asarray(data), axis), np.sqrt(N))

def take_equally_spaced(arr, n):
    """Take n equally spaced elements from arr
    
    We avoid the endpoints. So, divide the array into n+1 segments, and
    take the highest point of each segment, discarding the last.
    """
    # e.g., we want 2 equally spaced, so they are at 1/3 and 2/3
    arr = np.asarray(arr)
    first_element_relative = old_div(1.0, (n + 1))
    relative_pos = np.linspace(
        first_element_relative, 1 - first_element_relative, n)
    absolute_pos = my.rint((len(arr) - 1) * relative_pos)
    return arr[absolute_pos]

def define_integer_bin_edges(start, stop, n_bins=None, binwidth=None,
    error_on_uneven_bins=True):
    """Return integer bin edges from start to stop
    
    start : start of first bin
    stop : stop of last bin
    
    Either n_bins or binwidth should be provided, but not both.
    
    If n_bins is provided but the bin edges do not fall on integers, they
    will be rounded off, which will result in uneven bins. 
    An exception is raised if error_on_uneven_bins.
    
    If binwidth is provided but the range does not divide evenly, an
    exception is raised.
    
    Returns: bin edges
        The first will be start and the last will be stop.
        Each value will be an integer.
    """
    # Error check
    if stop < start:
        raise ValueError("stop must be greater than start")
    
    # Convert start and stop to integers
    if start != int(start):
        raise ValueError("start is not an integer")
    if stop != int(stop):
        raise ValueError("start is not an integer")
    start = int(start)
    stop = int(stop)
    
    # Operation depends on which parameter was provided
    if n_bins is not None and binwidth is None:
        # number of bins was provided
        if n_bins <= 0:
            raise ValueError("n_bins must be non-negative")
        if int(n_bins) != n_bins:
            raise ValueError("n_bins must be integer")

        if np.mod(stop - start, n_bins) != 0:
            # Does not divide evenly
            if error_on_uneven_bins:
                raise ValueError("specified n_bins does not divide evenly")
            
            # Do float and then round off
            fres = np.linspace(start, stop, n_bins)
            res = np.round(fres).astype(np.int)
        else:
            # Divides evenly so use arange
            binwidth = (stop - start) // n_bins
            res = np.arange(start, stop + 1, binwidth, dtype=np.int)
        
    elif binwidth is not None and n_bins is None:
        # bin width was provided
        if binwidth <= 0:
            raise ValueError("binwidth must be non-negative")
        if int(binwidth) != binwidth:
            raise ValueError("binwidth must be integer")
        
        res = np.arange(start, stop + 1, binwidth, dtype=np.int)
        if res[-1] != stop:
            raise ValueError("specified bin width does not divide evenly")
    
    else:
        raise ValueError("specify n_bins or binwidth but not both")

    return res

def times_near_times(x, y, dstart, dstop, sort_x=True):
    """Identify times in 'y' that are within [dstart, dstop) of times in x
    
    The algorithm is: does adding dstop to y, versus adding dstart to y,
    change the way it is sorted within x? If so, then it must be close enough
    to some value in x that it jumps over it.
    
    x : times to compare to, e.g., times of "bad events"
    y : probe times, e.g., data
        We want to filter y and include only values in y that are far from
        bad events.
    dstart, dstop : times relatives to the times in x.
        These should be integers. Ideally they would be variable size
        intervals, but this messes up searchsorted.
    
    Returns: boolean array, same shape as y
        res[n] is True if 
            there is some value x' in x such that 
            y[n] is in the range [x' + dstart, x' + dstop]
    
    Example:
    bad_times = np.array([0, 5, 15])
    probe_times = np.array([1, 6, 11, 16])
    
    # All except the 3rd are within [-2, 2)
    In [152]: times_near_times(bad_times, probe_times, -2, 2)
    Out[152]: array([ True,  True, False,  True], dtype=bool)
    
    # This interval is impossible
    In [153]: times_near_times(bad_times, probe_times, 2, -2)
    Out[153]: array([False, False, False, False], dtype=bool)    
    
    # Because it's half-open, none are within this interval
    In [165]: times_near_times(bad_times, probe_times, 0, 1)
    Out[165]: array([False, False, False, False], dtype=bool)
    
    # Because it's half-open, it will trigger on the interval start
    In [188]: times_near_times(bad_times + 1, probe_times, 0, 1)
    Out[188]: array([ True,  True, False,  True], dtype=bool)
    """
    # Convert to array
    x = np.asarray(x)
    y = np.asarray(y)
    
    # Sort x if necessary
    if sort_x:
        x = np.sort(x)
    
    # Use 'right' to make the half-open interval work out correctly
    v0 = np.searchsorted(x + dstart, y, side='right')
    v1 = np.searchsorted(x + dstop, y, side='right')
    
    # Compare v0 and v1 to get the "sign" correct (e.g., if dstop < dstart,
    # nothing will ever trigger)
    return v0 > v1

def find_interval(event_times, interval_starts, interval_stops, 
    stop_type='open'):
    """Find the interval containing the events
    
    event_times : array-like
        An array of event times
    interval_starts, interval_stops : array-like
        The stops and starts of each intervals. They should be of equal
        length and they should be sorted. Not sure what happens if the
        intervals overlap.
    stop_type : 'open' or 'closed'
        If the intervals are half-open (Pythonic), specify 'open'.
        Otherwise if they are closed, specify 'closed'.
    
    Returns: array of float
        The index of the interval corresponding to each event, or nan if 
        no such interval
    
    Examples:
    Equally spaced points every 0.5 from -.5 to 5.0
    data = np.arange(-.5, 5.5, .5)
    Consider two intervals: [0, 2) and [2, 4)
    
    0, 0.5, 1, and 1.5 are in the first interval,
    2, 2.5, 3, 3.5 are in the second interval,
    and the rest are in no interval
    
    list(zip(data, find_interval(data, [0, 2], [2, 4])))
    [(-0.5, nan),
     (0.0, 0.0),
     (0.5, 0.0),
     (1.0, 0.0),
     (1.5, 0.0),
     (2.0, 1.0),
     (2.5, 1.0),
     (3.0, 1.0),
     (3.5, 1.0),
     (4.0, nan),
     (4.5, nan),
     (5.0, nan)]

    If the intervals are closed [0, 2], and [3, 4]
    list(zip(data, find_interval(data, [0, 3], [2, 4], stop_type='closed')))
    [(-0.5, nan),
     (0.0, 0.0),
     (0.5, 0.0),
     (1.0, 0.0),
     (1.5, 0.0),
     (2.0, 0.0),
     (2.5, nan),
     (3.0, 1.0),
     (3.5, 1.0),
     (4.0, 1.0),
     (4.5, nan),
     (5.0, nan)]
    """
    # Datatype
    event_times = np.asarray(event_times)
    interval_starts = np.asarray(interval_starts)
    interval_stops = np.asarray(interval_stops)
    
    # Error check
    if not (np.sort(interval_starts) == interval_starts).all():
        raise ValueError("interval_starts must be sorted")
    if not (np.sort(interval_stops) == interval_stops).all():
        raise ValueError("interval_stops must be sorted")
    
    # Note on searchsorted
    # When an event time is equal to an interval time, then side='right'
    # returns a bigger index than side='left'. That's what we typically want
    # Pythonically, because as soon as equality is achieved the event 
    # enters the next interval.
    
    # For each event, the index of the interval start that it is 
    # equal to or after
    start_idx = np.searchsorted(interval_starts, event_times, side='right') - 1

    # The corresponding interval stop
    if stop_type == 'open':
        # For each event, the index of the interval stop that it is 
        # strictly before
        stop_idx = np.searchsorted(interval_stops, event_times, side='right')
    elif stop_type == 'closed':
        # For each event, the index of the interval stop that it is 
        # equal to or strictly before
        stop_idx = np.searchsorted(interval_stops, event_times, side='left')

    # Where ss1 == ss2, the contact is within the inclusive interval
    # defined by some corresponding cycle_start and cycle_stop
    contained_mask = start_idx == stop_idx
    
    # Blank out interval where not contained
    res = start_idx.astype(float)
    res[~contained_mask] = np.nan
    
    return res

def cut_dataframe(df, column, edges, new_column='bin', dropna=True, **kwargs):
    """Bin dataframe by values in specified column
    
    column : name of column to bin
    edges : edges of bins
    new_columns : name of new column to add with integer bin labels
    kwargs : passed to pandas.cut
        Suggestion: right=False to make the bins include the left edge
        instead of the right edge
    
    TODO: allow adding an extra bin edge at the end
    
    pandas.cut does the binning, so the number of bins is len(edges) - 1.
    rows outside the edges are given NaN and dropped if dropna is True.
    """
    df = df.copy()
    
    # Insert column
    df[new_column] = pandas.cut(df[column], bins=edges, labels=False, **kwargs)
    
    # Drop outside range
    if dropna:
        df = df[~df[new_column].isnull()]
        df[new_column] = df[new_column].astype(np.int)
    
    return df

def interpolate_dataframe(df, axis=0):
    """Proper interpolation without extrapolation of dataframe
    
    http://stackoverflow.com/questions/25255496/dataframe-interpolate-extrapolates-over-trailing-missing-data
    """
    df_mask = df.fillna(method='backfill', axis=axis).isnull()
    res = df.interpolate(axis=axis)
    res[df_mask] = np.nan    
    return res

def load_matlab_csv(filename):
    """Read CSV written by matlab tablewrite into DataFrames
    
    Each entry in the table can be a scalar or a variable length array.
    If it is a variable length array, then Matlab generates a set of
    columns, long enough to hold the longest array. These columns have
    the variable name with an index appended.
    
    This function infers which entries are scalars and which are arrays.
    Arrays are grouped together and sorted by their index.
    
    Returns: scalar_df, array_df
        scalar_df : DataFrame of scalar values from the table
        array_df : DataFrame with MultiIndex on columns
            The first level is the array name
            The second level is the index within that array
    """
    # Read the CSV file
    tdf = pandas.read_table(filename, sep=',')
    cols = list(tdf.columns)

    # Figure out which columns correspond to scalars and which to arrays
    scalar_cols = [] # scalar column names
    arr_cols = [] # array column names, without index
    arrname2idxs = {} # dict of array column name to list of integer indices
    arrname2colnames = {} # dict of array column name to list of full names
    
    # Iterate over columns
    for col in cols:
        # If the name ends in "_" plus space plus digits, it's probably
        # from an array
        if col[-1] in '0123456789' and '_' in col:
            # Array col
            # Infer the array name and index
            colsplit = col.split('_')
            arr_idx = int(colsplit[-1])
            arr_name = '_'.join(colsplit[:-1])
            
            # Store
            if arr_name in arrname2idxs:
                arrname2idxs[arr_name].append(arr_idx)
                arrname2colnames[arr_name].append(col)
            else:
                arrname2idxs[arr_name] = [arr_idx]
                arrname2colnames[arr_name] = [col]
                arr_cols.append(arr_name)
        
        else:
            # Scalar col
            scalar_cols.append(col)

    # Extract all scalar columns
    scalar_df = tdf[scalar_cols]

    # Extract each set of array columns into its own dataframe
    array_df_d = {}
    for arrname in arr_cols:
        adf = tdf[arrname2colnames[arrname]].copy()
        adf.columns = arrname2idxs[arrname]
        array_df_d[arrname] = adf
    
    # Concatenate array dataframes
    array_df = pandas.concat(array_df_d, axis=1)

    return scalar_df, array_df

def sort_whisker_names(curated_whiskers_as_dict):
    """Sort whisker names in anatomical order.
    
    curated_whiskers_as_dict : dict from object label to whisker name
    
    Iterates through all values of `curated_whiskers_as_dict`. Sorts them
    into junk (anything containing "junk"), unk (anything containing "unk"),
    greeks (lowercase names), and real whiskers (the rest).
    
    Returns: dict
        'greek', 'real', 'junk', 'unk' : as above
        'sorted_order' : greek + real + junk + unk
    """
    # Order: real, then junk, then unk
    junk_whiskers, unk_whiskers, greek_whiskers, real_whiskers = [], [], [], []
    for whisker_name in list(curated_whiskers_as_dict.values()):
        if 'junk' in whisker_name:
            junk_whiskers.append(whisker_name)
        elif 'unk' in whisker_name:
            unk_whiskers.append(whisker_name)
        elif str.islower(whisker_name):
            greek_whiskers.append(whisker_name)
        else:
            real_whiskers.append(whisker_name)
    sorted_whisker_names = (sorted(greek_whiskers) + sorted(real_whiskers) + 
        sorted(junk_whiskers) + sorted(unk_whiskers))
    
    return {
        'greek': greek_whiskers,
        'junk': junk_whiskers,
        'unk': unk_whiskers,
        'real': real_whiskers,
        'sorted_order': sorted_whisker_names,
    }
    
def simple_sort_whisker_names(whisker_names):
    # Order: real, then junk, then unk
    junk_whiskers, unk_whiskers, greek_whiskers, real_whiskers = [], [], [], []
    
    for whisker_name in whisker_names:
        if 'junk' in whisker_name:
            junk_whiskers.append(whisker_name)
        elif 'unk' in whisker_name:
            unk_whiskers.append(whisker_name)
        elif str.islower(whisker_name):
            greek_whiskers.append(whisker_name)
        else:
            real_whiskers.append(whisker_name)
    sorted_whisker_names = (sorted(greek_whiskers) + sorted(real_whiskers) + 
        sorted(junk_whiskers) + sorted(unk_whiskers))
    
    return sorted_whisker_names

def insert_level(df, func, name, level=0, sort=True):
    """Inserts a level with the mouse name, using a level with the session
    
    level : integer position of new mouse level
    
    sort : bool
        Whether to sort before returning
    
    Returns : DataFrame with new level on index
    """
    # Copy without changing
    df = df.copy()
    
    # Get the index
    idx = df.index.to_frame().reset_index(drop=True)
    
    # Calculate new level
    to_insert = func(idx)
    
    # Insert the mouse column
    idx.insert(level, name, to_insert)
    
    # Reset the index
    df.index = pandas.MultiIndex.from_frame(idx)
    
    # Sort
    df.sort_index(inplace=True)
    
    return df

def insert_mouse_level(df, level=0, sort=True):
    """Inserts a level with the mouse name, using a level with the session
    
    level : integer position of new mouse level
    
    sort : bool
        Whether to sort before returning
    
    Returns : DataFrame with new level on index
    """
    if 'session' not in df.index.names:
        raise ValueError("DataFrame must have session as a level on the index")
    
    def session_name2mouse_name(idx_df):
        return [session.split('_')[1] for session in idx_df['session']]
    
    res = insert_level(
        df, func=session_name2mouse_name, name='mouse', 
        level=level, sort=sort)
    
    return res

def insert_mouse_and_task_levels(df, mouse2task, level=0, sort=True):
    """Insert levels for mouse and task
    
    df : DataFrame
    
    mouse2task : Series mapping mouse name to task name
    
    Returns : DataFrame
        A copy of `df` with the new levels inserted
    """
    # Insert mouse level
    df = wwutils.misc.insert_mouse_level(df)
    
    # Insert task level
    df = wwutils.misc.insert_level(
        df, name='task', 
        func=lambda idx: idx['mouse'].map(mouse2task))    
    
    return df

def slice_df_by_some_levels(df, slicing_midx):
    """Slice DataFrame by a subset of the levels in its index.
    
    df : DataFrame
    
    slicing_midx : a MultiIndex to slice it with
    
    Returns : DataFrame
    """
    # The levels to slice on, in sorted order
    slicing_levels = list(slicing_midx.names)
    
    # The levels not to slice on
    non_slicing_levels = [level for level in df.index.names 
        if level not in slicing_levels]
    
    # Error check
    if not np.in1d(slicing_midx.names, df.index.names).all():
        raise ValueError("cannot slice on missing levels")
    
    # Convert slicing_midx to DataFrame
    df1 = slicing_midx.to_frame().reset_index(drop=True)
    
    # Drop duplicates on the slicing midx (otherwise merging makes it too big)
    df1 = df1.drop_duplicates()
    
    df2 = df.index.to_frame().reset_index(drop=True)
    df1['key'] = 1
    mask = ~pandas.merge(
        df2, df1, on=slicing_levels, how='left')[
        'key'].isnull()
    return df.loc[mask.values]    

def assert_index_equal_on_levels(df1, df2, levels):
    """Check that the indexes of df1 and df2 are equal, considering only levels
    
    """
    
    # Get sorted unique values of `levels` on `df1`
    df1_levels = df1.index.to_frame()[levels].reset_index( 
        drop=True).drop_duplicates().sort_values(levels).reset_index(
        drop=True)
    
    # Same for df2
    df2_levels = df2.index.to_frame()[levels].reset_index( 
        drop=True).drop_duplicates().sort_values(levels).reset_index(
        drop=True)
    
    # Assert
    assert df1_levels.equals(df2_levels)

def find_image_shift(img0, img1):
    """Find the best shift to make img0 look like img1
    
    img0 : 2d reference image
    img1 : 2d test image
    
    Returns: numpy array of shape (2,)
        This is the displacement to apply to img0 to make it look like img1
        It is in (dim0, dim1) / (r, c) / (y, x) order
        scipy.ndimage.shift(img0, displacement) will be approx equal to img1
        scipy.ndimage.shift(img1, -displacement) will be approx equal to img0
    """
    # Use this shape to interpret the offset
    image_dims = img0.shape
    half_image_dims = (image_dims[0] // 2, image_dims[1] // 2)
    
    # we have to reverse the second image to make it a true "convolution"
    img_corr = scipy.signal.fftconvolve(img1, img0[::-1,::-1], 
        mode="same")
    
    # Find the peak (2d offset of best correlation)
    brightest_spot = np.unravel_index(
        np.argmax(img_corr), img_corr.shape)

    # Convert to a displacement. A peak in the center means no displacement.
    # These are in (dim0, dim1) / (r, c) / (y, x) order.
    displacement = np.array([
        brightest_spot[0] - half_image_dims[0], 
        brightest_spot[1] - half_image_dims[1],
        ])    
    
    #~ # Debug plot
    #~ f, axa = plt.subplots(2, 2)
    #~ wwutils.plot.imshow(img0, ax=axa[0, 0])
    #~ axa[0, 0].set_title('img0')
    #~ wwutils.plot.imshow(img1, ax=axa[0, 1])
    #~ axa[0, 1].set_title('img1')
    #~ wwutils.plot.imshow(scipy.ndimage.shift(img0, displacement), ax=axa[1, 1])
    #~ axa[1, 1].set_title('img0 shifted to look like img1')
    #~ wwutils.plot.imshow(scipy.ndimage.shift(img1, -displacement), ax=axa[1, 0])
    #~ axa[1, 0].set_title('img1 shifted to look like img0')
    
    return displacement

def CustomCmap(from_rgb,to_rgb):
    from matplotlib.colors import LinearSegmentedColormap

    # https://stackoverflow.com/questions/16267143/matplotlib-single-colored-colormap-with-saturation

    # from color r,g,b
    r1,g1,b1 = from_rgb

    # to color r,g,b
    r2,g2,b2 = to_rgb

    cdict = {'red': ((0, r1, r1),
                   (1, r2, r2)),
           'green': ((0, g1, g1),
                    (1, g2, g2)),
           'blue': ((0, b1, b1),
                   (1, b2, b2))}

    cmap = LinearSegmentedColormap('custom_cmap', cdict)
    return cmap

def transform(data, transformation_df):
    """Transform coordinates in `data` using `transformation_df`
    
    data : 2d DataFrame with 2 columns, ending in 'x' and 'y'
    transformation_df : must have columns shift_x, shift_y, c00, c01, c10, c11
    
    Returns: DataFrame shaped like `data`    
    """
    assert data.shape[1] == 2
    
    # Figure out which column is x and which is y
    x_col = None
    y_col = None
    for colname in data.columns:
        if colname.endswith('x'):
            x_col = colname
        if colname.endswith('y'):
            y_col = colname
    assert x_col is not None
    assert y_col is not None
    
    # Rename x and y
    data2 = data.rename(columns={x_col: 'x', y_col: 'y'})
    
    # Apply the shift
    data2 = data2 + transformation_df[['shift_x', 'shift_y']].rename(
        columns={'shift_x': 'x', 'shift_y': 'y'})
    
    # Apply the rotation and scaling
    res_x = transformation_df['c00'] * data2['x'] + transformation_df['c01'] * data2['y']
    res_y = transformation_df['c10'] * data2['x'] + transformation_df['c11'] * data2['y']
    
    # Concat 
    res = pandas.concat([res_x.rename(x_col), res_y.rename(y_col)], axis=1)
    
    # Reindex like input data
    res = res.reindex(data.columns, axis=1)
    
    # Return
    return res

def fetch_columns_with_error_check(df, columns):
    """Slice columns from df and error check that they all exist"""
    # Extract
    res = df.loc[:, columns].copy()
    res.columns = res.columns.remove_unused_levels()

    # Error check
    # A missing label will be silently ignored above
    assert sorted(res.columns.levels[0]) == sorted(columns)    
    
    return res

def create_dir_if_does_not_exist(dirname):
    if not os.path.exists(dirname):
        os.mkdir(dirname)

def gaussian_sum_smooth(xdata, ydata, xeval, sigma, null_thresh=0.6):
    """Apply gaussian sum filter to data.
    
    xdata, ydata : array
        Arrays of x- and y-coordinates of data. 
        Must be 1d and have the same length.
    
    xeval : array
        Array of x-coordinates at which to evaluate the smoothed result
    
    sigma : float
        Standard deviation of the Gaussian to apply to each data point
        Larger values yield a smoother curve.
    
    null_thresh : float
        For evaluation points far from data points, the estimate will be
        based on very little data. If the total weight is below this threshold,
        return np.nan at this location. Zero means always return an estimate.
        The default of 0.6 corresponds to approximately one sigma away 
        from the nearest datapoint.
    """
    # Distance between every combination of xdata and xeval
    # each row corresponds to a value in xeval
    # each col corresponds to a value in xdata
    delta_x = xeval[:, None] - xdata

    # Calculate weight of every value in delta_x using Gaussian
    # Maximum weight is 1.0 where delta_x is 0
    weights = np.exp(-0.5 * ((delta_x / sigma) ** 2))

    # Multiply each weight by every data point, and sum over data points
    smoothed = np.dot(weights, ydata)

    # Nullify the result when the total weight is below threshold
    # This happens at evaluation points far from any data
    # 1-sigma away from a data point has a weight of ~0.6
    nan_mask = weights.sum(1) < .6
    smoothed[nan_mask] = np.nan
    
    # Normalize by dividing by the total weight at each evaluation point
    # Nullification above avoids divide by zero warning shere
    smoothed = smoothed / weights.sum(1)

    
    return smoothed
    
def gaussian_sum_smooth_pandas(data, evaluation_index, sigma, 
    null_thresh=0.6, drop_null=False):
    """Apply gaussian sum filter to Series.
    
    data : Series
    
    evaluation_index : array-like or index
        Points at which to evaluate the smoothed result.
    
    sigma, null_thresh : See gaussian_sum_smooth
    
    drop_null : if True, drop null values before returning
    
    
    Returns: Series
        The index are `evaluation_index`
        The values are the smoothed data
    """
    # Parse into arrays
    arr = gaussian_sum_smooth(
        data.index.values, data.values, 
        np.asarray(evaluation_index), 
        sigma=sigma, null_thresh=null_thresh,
        )
    
    # Convert to Series
    ser = pandas.Series(arr, index=evaluation_index)
    
    # Optionally drop null
    if drop_null:
        ser = ser.dropna()
    
    return ser

def load_bwid(*args, **kwargs):
    print("warning: replace wwutils.misc.load_bwid with wwutils.dataload.load_bwid")
    return wwutils.dataload.load_bwid(*args, **kwargs)
    
def stack_df_to_series(df):
    """Stack DataFrame until it becomes a series"""

    res = df.copy()
    for n_stack in range(data.columns.nlevels):
        res = res.stack()
    
    return res


""" mfile_io.py

Author: Nathan Clack <clackn@janelia.hhmi.org>  
Date  : 25 January 2009

Copyright 2010 Howard Hughes Medical Institute.
All rights reserved.
Use is subject to Janelia Farm Research Campus Software Copyright 1.1
license terms (http://license.janelia.org/license/jfrc_copyright_1_1.html).

Added to WhiskiWrap repository by Vincent Prevosto, 05/2023
"""
import os,sys
from ctypes import *
from ctypes.util import find_library
import numpy
from numpy import zeros, double, fabs, ndarray, array
from . import wfile_io
from .wfile_io import cWhisker_Seg

import warnings

import pdb
from functools import reduce

# Check if WHISKPATH is set in environment variables
whisk_bin_path = os.environ.get('WHISKPATH', None)

# if not whisk_bin_path:
#     # You can set a default path here if desired, or raise an error.
#     raise EnvironmentError("WHISKPATH environment variable not set!")

if sys.platform == 'win32':
    lib = os.path.join(whisk_bin_path, 'whisk.dll')
else:
    lib = os.path.join(whisk_bin_path, 'libwhisk.so')

# No need to append to PATH if you're directly specifying the path to the library.
name = find_library('whisk')
if not name:
    name = lib

try:
    cWhisk = cdll.LoadLibrary(name)
except:
    raise ImportError("Cannot load whisk shared library")

if cWhisk._name is None:
    raise ImportError("Cannot load whisk shared library")


# dllpath = os.path.split(os.path.abspath(__file__))[0]
# if sys.platform == 'win32':
#   lib = os.path.join(dllpath,'whisk.dll')
# else:
#   lib = os.path.join(dllpath,'libwhisk.so')
# os.environ['PATH']+=os.pathsep + os.pathsep.join(['.','..',dllpath])
# name = find_library('whisk')
# if not name:
#   name=lib
# try:
#   cWhisk = cdll.LoadLibrary( name )
# except:
#   raise ImportError("Can not load whisk shared library"); 
# if cWhisk._name==None:
#   raise ImportError("Can not load whisk shared library");

_param_file = "default.parameters"
if cWhisk.Load_Params_File(_param_file)==1: #returns 0 on success, 1 on failure
  raise Exception("Could not load tracing parameters from file: %s"%_param_file)

class cMeasurements(Structure):
  """ Proxy for Measurements struct. 
  >>> from numpy.random import rand
  >>> data = rand(20,10)
  >>> table = cWhisk.Measurements_Table_From_Doubles( data.ctypes.data_as(POINTER(c_double)), 20, 10 )
  >>> table[0].n
  7
  >>> table   # doctest:+ELLIPSIS
  <ctypes.LP_cMeasurements object at ...>
  """
  _fields_ = [("row",            c_int               ),
              ("fid",            c_int               ),                                                           
              ("wid",            c_int               ),                                                           
              ("state",          c_int               ),                                                           
              ("face_x",         c_int               ),             #// used in ordering whiskers on the face...roughly, the center of the face                                              
              ("face_y",         c_int               ),             #//                                      ...does not need to be in image                                                 
              ("col_follicle_x", c_int               ),             #// index of the column corresponding to the folicle x position                                                          
              ("col_follicle_y", c_int               ),             #// index of the column corresponding to the folicle y position                                                          
              ("valid_velocity", c_int               ),                                                           
              ("n",              c_int               ),                                                           
              ("face_axis",      c_char              ),
              ("data",           POINTER( c_double ) ),             # // array of n elements                      
              ("velocity",       POINTER( c_double ) )]             # // array of n elements - change in data/time

class cDistributions(Structure):
  """
  >>> this = cWhisk.Alloc_Distributions( 32, 8, 4 )
  >>> this                                          # doctest:+ELLIPSIS 
  <ctypes.LP_cDistributions object at ...>
  >>> cWhisk.Free_Distributions( this )
  """
  _fields_ = [("n_measures",   c_int               ),
              ("n_states",     c_int               ),
              ("n_bins",       c_int               ),
              ("bin_min",      POINTER( c_double ) ),   # // array of n_measures elements                                                          
              ("bin_delta",    POINTER( c_double ) ),   # // array of n_measures elements                                                          
              ("data",         POINTER( c_double ) )]   # // array of holding histogram information with dimensions (n_bins,n_measures,n_states)
  def asarray(self):
    d = zeros( (self.n_states, self.n_measures, self.n_bins) )
    cWhisk.Copy_Distribution_To_Doubles( byref(self), d.ctypes.data_as( POINTER(c_double) ) )
    return d
  
  def bins_as_array(self):
    b = zeros( (self.n_measures, self.n_bins) )
    cWhisk.Distributions_Bins_To_Doubles( byref(self), b.ctypes.data_as( POINTER(c_double) ) )
    return b

class MeasurementsTable(object):
  """
  >>> data = numpy.load('data/testing/seq140[autotraj].npy')
  >>> table = MeasurementsTable(data)
  >>> table._measurements[0].n
  8
  """
  def __init__(self, datasource):
    """
    Load table from numpy array or from a file.

    >>> table = MeasurementsTable( zeros((500,5)) )

    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )

    >>> import wfile_io
    >>> wvd = wfile_io.Load_Whiskers( "data/testing/seq140.whiskers" )
    >>> table = MeasurementsTable( {'whiskers':wvd, 'facehint':'left'} )

    """
    object.__init__(self)
    self._measurements = None
    self._nrows = 0
    self._sort_state = None
    self._free_measurements = cWhisk.Free_Measurements_Table
    if isinstance(datasource,str):
      self._load(datasource)
    elif isinstance(datasource,dict):
      wvd      = datasource['whiskers']
      facehint = datasource['facehint']
      self._measurements, self._nrows = MeasurementsTable._fromWhiskerDictWithFacehint( wvd, facehint )
    else:
      self._measurements = cWhisk.Measurements_Table_From_Doubles( 
                              datasource.ctypes.data_as( POINTER(c_double) ),   # data buffer       
                              datasource.shape[0],                              # number of rows    
                              datasource.shape[1] )                             # number of columns 
      self._nrows = datasource.shape[0]

  def __del__(self):
    """
    >>> table = MeasurementsTable( zeros((500,5)) )
    >>> del table
    """
    self._free_measurements(self._measurements)

  @staticmethod
  def _fromWhiskerDict(wvd, face_xy, faceaxis ):
    """
    Returns: LP_cMeasurements, int
      
    Warning: the returned cMeasurements object needs to be properly deallocated
             when finished.  Potential memory leak.  For this reason, use the 
             MeasurementsTable constructor (__init__) instead.
    """
    facex,facey = face_xy
    wv = wfile_io.cWhisker_Seg.CastDictToArray(wvd)
    return cWhisk.Whisker_Segments_Measure(wv,len(wv), facex, facey, faceaxis), len(wv)
  
  @staticmethod
  def _fromWhiskerDictWithFacehint(wvd, facehint ):
    """
    Returns: LP_cMeasurements, int
      
    Warning: the returned cMeasurements object needs to be properly deallocated
             when finished.  Potential memory leak.  For this reason, use the 
             MeasurementsTable constructor (__init__) instead.
    """
    x,y,ax = c_int(),c_int(),c_char()
    wv = wfile_io.cWhisker_Seg.CastDictToArray(wvd)
    cWhisk.face_point_from_hint( wv, len(wv), facehint, byref(x), byref(y), byref(ax))
    return cWhisk.Whisker_Segments_Measure(wv,len(wv), x.value, y.value, ax.value), len(wv)


  def asarray(self):
    """  
    >>> from numpy.random import rand
    >>> data = rand(200,10)
    >>> table = MeasurementsTable(data)
    >>> shape = table.asarray()
    >>> print shape.shape
    (200, 10)
    >>> print (shape[:,3:]==data[:,3:]).all()
    True
    """
    if self._nrows==0:
      return []
    data = zeros( (self._nrows, self._measurements[0].n+3), dtype=double )
    cWhisk.Measurements_Table_Data_To_Doubles(self._measurements, 
                                             self._nrows, 
                                             data.ctypes.data_as( POINTER( c_double ))
                                            );
    return data

  def get_trajectories(self):
    """
    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> traj  = table.get_trajectories()
    >>> max(traj.keys())
    3
    >>> traj.has_key(-1)
    False
    """
    data = self.asarray()
    t = {}
    for row in data:
      r = list(map(int,row[:3]))
      t.setdefault( r[0],{} ).setdefault( r[1], r[2] ) 
    if -1 in list(t.keys()):
      del t[-1]
    return t

  def save_trajectories(self, filename, excludes=[]):
    """  Saves to a trajectories file.

    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> table.save_trajectories( "data/testing/trash.trajectories" ) # doctest:+ELLIPSIS 
    <...MeasurementsTable object at ...>
    """
    trajectories = self.get_trajectories()
    f = open( filename, 'w' )
    for k,v in trajectories.items():
      if not k in excludes:
        for s,t in v.items():
          print('%d,%d,%d'%(k,s,t), file=f)
    return self

  def load_trajectories(self,filename ):
    """  Loads trajectories and saves them to the table.
    Trajectory id's correspond to the `state` label.

    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> table.save_trajectories( "data/testing/trash.trajectories" )  # doctest:+ELLIPSIS 
    <...MeasurementsTable object at ...>
    >>> table.load_trajectories( "data/testing/trash.trajectories" )  # doctest:+ELLIPSIS 
    <...MeasurementsTable object at ...>
    """
    trajectories = {}
    f = open( filename, 'r' )

    cur = 0;
    for line in f:
      t = [int(x) for x in line.split(',')[:3]]
      if not t[0] in trajectories:
        trajectories[t[0]] = {}
      trajectories[ t[0] ][ t[1] ] = t[2];

    self.commit_trajectories( trajectories )
    return self

  def commit_trajectories(self,traj):
    """
    >>> traj = {0: {0:0,1:0}, 1: {0:1,1:1} }
    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> mn,mx = table.commit_trajectories(traj).get_state_range()
    >>> mn
    0
    >>> mx
    1
    >>> lentraj = lambda x: len(table.get_shape_data(x))
    >>> lentraj(0)
    2
    >>> add = lambda a,b:a+b
    >>> table._nrows == reduce(add, map(lentraj,xrange(mn-1,mx+1)))
    True
    """
    inv = {}
    for tid,t in traj.items():
      for k in t.items():
        inv[k] = tid  

    for i in range(self._nrows):  #update new
      row = self._measurements[i]
      s = inv.get( (row.fid,row.wid) )
      row.state = s if (not s is None) else -1 

    return self

  def get_state_range(self):
    """
    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> mn,mx = MeasurementsTable(data).update_velocities().get_state_range()
    >>> mn
    0
    >>> mx
    3
    """ 
    mn,mx = c_int(),c_int()
    sorted = (not self._sort_state is None ) and \
             ("state" in self._sort_state  )
    n = cWhisk._count_n_states(self._measurements,
                              self._nrows,
                              sorted,
                              byref(mn),
                              byref(mx))
    f = lambda x: x.value if x.value >=0 else 0
    return list(map(f,[mn,mx]))

  def iter_state(self):
    """
    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> for i in table.update_velocities().iter_state():
    ...     print i
    ...     
    0
    1
    2
    3
    """
    mn,mx = self.get_state_range()
    return range(mn,mx+1)

  def get_shape_table(self):
    """  
    >>> from numpy.random import rand
    >>> data = rand(200,10)
    >>> table = MeasurementsTable(data)
    >>> shape = table.get_shape_table()
    """
    shape = zeros( (self._nrows, self._measurements[0].n), dtype=double )
    cWhisk.Measurements_Table_Copy_Shape_Data( self._measurements, 
                                              self._nrows, 
                                              shape.ctypes.data_as( POINTER(c_double) ) )
    return shape

  def get_time_and_mask(self, state, rows = None):
    """
    Returns `time` and `valid velocity` mask for selected state.  
    Order of results is determined by the table's sort order.

    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).update_velocities()
    >>> time,mask = table.get_time_and_mask(1)
    """
    if rows is None:
      rows = cWhisk.Measurements_Table_Size_Select_State( self._measurements, self._nrows, int(state) )
    time = zeros( rows, dtype = double )
    mask = zeros( rows, dtype = int    )
    cWhisk.Measurements_Table_Select_Time_And_Mask_By_State( self._measurements, 
                                                            self._nrows,
                                                            int(state),
                                                            time.ctypes.data_as (POINTER( c_double )), 
                                                            mask.ctypes.data_as (POINTER( c_int    )) )
    return time,mask

  def get_velocities(self, state, rows = None):
    """
    Returns velocity for selected state.  
    Order of results is determined by the table's sort order.

    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data)
    >>> table.update_velocities() # doctest:+ELLIPSIS 
    <...MeasurementsTable object at ...>
    >>> velocities = table.get_velocities(1)
    """
    if rows is None:
      rows = cWhisk.Measurements_Table_Size_Select_State( self._measurements, self._nrows, int(state) )
    vel  = zeros( (rows, self._measurements[0].n ), dtype = double )
    cWhisk.Measurements_Table_Select_Velocities_By_State( self._measurements, 
                                                         self._nrows,
                                                         int(state),
                                                         vel.ctypes.data_as (POINTER( c_double )) )
    return vel
  
  def get_shape_data(self, state, rows = None):
    """
    Returns shape data for selected state.  
    Order of results is determined by the table's sort order.

    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).update_velocities()
    >>> shape = table.get_shape_data(1)
    
    >>> table = MeasurementsTable('data/testing/seq140[autotraj].measurements').update_velocities()
    >>> shape = table.get_shape_data(1)
    """
    if rows is None:
      rows = cWhisk.Measurements_Table_Size_Select_State( self._measurements, self._nrows, int(state) )
    data  = zeros( (rows, self._measurements[0].n ), dtype = double )
    cWhisk.Measurements_Table_Select_Shape_By_State( self._measurements, 
                                                    self._nrows,
                                                    int(state),
                                                    data.ctypes.data_as (POINTER( c_double )) )
    return data

  def get_data(self, state, rows = None ):
    """
    Returns time, shape, velocity and velocity_valid  data for selected state.  
    Order of results is determined by the table's sort order.

    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).update_velocities()
    >>> time,shp,vel,mask = table.get_data(1)
    """
    if rows is None:
      time,mask = self.get_time_and_mask(state) 
    else:
      time,mask = self.get_time_and_mask(state, rows = rows) 
    vel = self.get_velocities(state, rows = time.shape[0] )
    shp = self.get_shape_data(state, rows = time.shape[0] )
    return time, shp, vel, mask

  def get_velocities_table(self):
    """
    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).update_velocities()
    >>> vel = table.get_velocities_table()
    """
    vel = zeros( (self._nrows, self._measurements[0].n), dtype=double )
    cWhisk.Measurements_Table_Copy_Velocities( self._measurements, 
                                              self._nrows, 
                                              vel.ctypes.data_as( POINTER(c_double) ) )
    return vel

  def set_constant_face_position(self, x, y):
    """
    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> table = table.set_constant_face_position( -100, 100 )
    >>> table._measurements[0].face_x
    -100
    >>> table._measurements[0].face_y
    100
    """
    cWhisk.Measurements_Table_Set_Constant_Face_Position( self._measurements, self._nrows, x, y )
    return self

  def set_follicle_position_column(self, ix, iy):
    """
    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> table = table.set_follicle_position_column( 7, 8 ) 
    >>> table._measurements[0].col_follicle_x
    7
    >>> table._measurements[0].col_follicle_y
    8
    """
    cWhisk.Measurements_Table_Set_Follicle_Position_Indices( self._measurements, self._nrows, ix, iy )
    return self

  def sort_by_state_time(self):
    """
    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).sort_by_state_time() 
    >>> table._measurements[0].state
    -1
    >>> table._measurements[0].fid
    0
    >>> table._measurements[table._nrows].state
    0
    >>> table._measurements[table._nrows-1].state
    3
    >>> table._measurements[table._nrows-1].fid
    4598
    >>> table._sort_state
    'state,time'
    """
    sortstate = "state,time"
    if self._sort_state != sortstate:
      cWhisk.Sort_Measurements_Table_State_Time( self._measurements, self._nrows )
      self._sort_state = sortstate
    return self

  def sort_by_time(self):
    """
    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).sort_by_time() 
    >>> table._measurements[0].fid
    0
    >>> table._measurements[table._nrows-1].fid
    4598
    >>> table._sort_state
    'time'
    """
    sortstate = "time"
    if self._sort_state != sortstate:
      cWhisk.Sort_Measurements_Table_Time( self._measurements, self._nrows )
      self._sort_state = sortstate
    return self

  def sort_by_time_face(self):
    """
    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> table = table.set_constant_face_position( -100, 100 ).set_follicle_position_column( 7, 8 )
    >>> table = table.sort_by_time_face()
    >>> table._measurements[0].fid
    0
    >>> table._measurements[table._nrows-1].fid
    4598
    >>> table._sort_state
    'time,face'
    """
    sortstate = "time,face"
    if(self._sort_state != sortstate):
      cWhisk.Sort_Measurements_Table_Time_Face( self._measurements, self._nrows )
      self._sort_state = sortstate
    return self

  def update_velocities(self):
    """
    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).update_velocities() 
    >>> vel = table.get_velocities_table()
    """
    self.sort_by_state_time()
    cWhisk.Measurements_Table_Compute_Velocities( self._measurements, self._nrows )
    return self

  def save(self, filename):
    """
    >>> data = numpy.load('data/testing/seq140[autotraj].npy')
    >>> table = MeasurementsTable(data).update_velocities()
    >>> table.save( "data/testing/trash.measurements" )    # doctest:+ELLIPSIS 
    <...MeasurementsTable object at ...>
    """
    cWhisk.Measurements_Table_To_Filename( filename, None, self._measurements, self._nrows )
    return self

  def save_to_matlab_file(self, filename, format = '5'):
    """
    Saves shape measurements to Matlab's .mat format.

    This uses the `scipy.io.matlab.savemat` function.  See that functions documentation for
    details on input options.

    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> table.save_to_matlab_file( "data/testing/trash.mat" )    # doctest:+ELLIPSIS 
    <...MeasurementsTable object at ...>
    """
    from scipy.io.matlab import savemat
    kwargs = locals().copy()
    for k in [ 'self', 'savemat', 'filename' ]:
      del kwargs[k]
    savemat( filename, 
            { 'measurements': self.asarray() }, 
            **kwargs)
    return self

  def _load(self, filename):
    """
    Loads table from a saved file.

    >>> table = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    """
    if not os.path.exists(filename):
      raise IOError("Could not find file %s"%filename)
    nrows = c_int()
    if self._measurements:
      cWhisk.Free_Measurements_Table( self._measurements )

    # import ctypes

    # Define the argument types for Measurements_Table_From_Filename
    # cWhisk.Measurements_Table_From_Filename.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_int)]

    # Prepare the filename and other variables
    bfilename = bytes(filename, 'utf-8') # Convert the string to a bytes object
    formats = [bytes(f, 'utf-8') for f in ['v0', 'v1', 'v2', 'v3']]
    # See Measurements_File_Formats in measurements_io.c for more information. v3 is default and will be 
    # automatically detected anyway. Older formats are for 32bit systems. 

    self._measurements = cWhisk.Measurements_Table_From_Filename(bfilename, formats[-1], byref(nrows))
                                                    #  ctypes.byref(nrows))
    self._nrows = nrows.value
    self._sort_state = None #unknown
    return self
  
  def diff_identity(self, table):
    """
    Searches two tables for different identity assignments and returns
    a list of frames where a difference was found.  Ideally, the two
    tables would have been derived from the same movie.

    If the tables are identical, an empty list is returned:

    >>> A = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> B = MeasurementsTable( "data/testing/seq140[autotraj].measurements" )
    >>> A.diff_identity(B)
    []

    The differences are not symmetric.  This is because "junk" states in the
    table on the left-hand side (`self`) are ignored.

    >>> B = MeasurementsTable( "data/testing/seq140[solve].measurements" )
    >>> len(B.diff_identity(A))
    69
    >>> len(A.diff_identity(B))
    25
    """
    nframes = c_int(0)
    frames = cWhisk.Measurements_Tables_Get_Diff_Frames( self._measurements, self._nrows, 
                                                        table._measurements, table._nrows, 
                                                        byref(nframes) )
    return [frames[i] for i in range(nframes.value)]

  def est_length_threshold(self,lowpx=1.0/0.04,highpx=50.0/0.04):
    ncount = c_int(0)
    thresh = cWhisk.Measurements_Table_Estimate_Best_Threshold(
        self._measurements,
        self._nrows,
        c_int(0),           # length column
        lowpx,highpx,
        1,                  # use greater than
        byref(ncount))      # estimated number of whiskers
    return thresh,ncount
        
class Distributions(object):
  def __init__(self, table = None, nbins = 32):
    """
    Create an empty Distributions object:
    
    >>> dists = Distributions()
    
    Initialize Distributions using a MeasurementTable:
    
    >>> import numpy
    >>> data = numpy.load( "data/testing/seq140[autotraj].npy" )
    >>> table = MeasurementsTable(data)

    >>> table = MeasurementsTable('data/testing/seq140[autotraj].measurements')
    >>> dists = Distributions(table.update_velocities())         # doctest:+ELLIPSIS  
    ...
    """
    object.__init__(self)
    self._free_distributions = cWhisk.Free_Distributions
    self._shp = None
    self._vel = None
    if not table is None:
      self.build(table, nbins)  
  
  def __del__(self):
    if self._shp:
      self._free_distributions( self._shp )
    if self._vel:
      self._free_distributions( self._vel )
  
  def build(self, table, nbins = 32):
    """
    >>> dists = Distributions()
    >>> table = MeasurementsTable('data/testing/seq140[autotraj].measurements')
    >>> dists.build(table)                   # doctest:+ELLIPSIS  
    <...Distributions object at ...>
    """
    assert isinstance(table,MeasurementsTable), "Wrong type for table."
    table.sort_by_state_time()
    self._shp = cWhisk.Build_Distributions         ( table._measurements, table._nrows, nbins )
    cWhisk.Distributions_Normalize( self._shp )
    cWhisk.Distributions_Apply_Log2( self._shp )
    table.update_velocities()
    self._vel = cWhisk.Build_Velocity_Distributions( table._measurements, table._nrows, nbins ) #changes table's sort order
    cWhisk.Distributions_Normalize( self._vel )
    cWhisk.Distributions_Apply_Log2( self._vel )
    table._sort_state = 'time'
    return self
  
  def velocities(self):
    """
    >>> dists = Distributions( MeasurementsTable('data/testing/seq140[autotraj].measurements') )
    >>> vbins, v = dists.velocities()
    """
    return self._vel[0].bins_as_array(), self._vel[0].asarray()  
  
  def shapes(self):
    """
    >>> dists = Distributions( MeasurementsTable('data/testing/seq140[autotraj].measurements') )
    >>> sbins, s = dists.shapes()
    """
    return self._shp[0].bins_as_array(), self._shp[0].asarray()  

# def solve( table ):
#   cWhisk.Solve( table._measurements, table._nrows, 32, 8096 )
#   table._sort_state = "time"
#   return table

def batch_make_measurements(sourcepath, ext = '*.seq', label = 'curated'):
  """
  To update/remake a measurements table, delete the *.npy and *.measurements
  files in the `sourcepath`.
  """
  warnings.simplefilter("ignore")
  from glob import glob
  from ui.whiskerdata import load_trajectories
  from .wfile_io import Load_Whiskers
  from . import summary
  warnings.simplefilter("default")

  def get_summary_data( filename, whiskers, trajectories ):
    if os.path.exists(filename):
      data = numpy.load(filename)
    else:
      data = array(list( summary.features(whiskers) ))
      numpy.save( filename, data )
      return data
    return summary.commit_traj_to_data_table( trajectories, data )

  for name in glob( os.path.join( sourcepath, ext ) ):
    root,ext = os.path.splitext( name )
    prefix = root + '[%s]'%label
    if not os.path.exists( prefix + '.measurements' ):
      t,tid = load_trajectories( prefix + '.trajectories' )
      print(prefix)
      print(list(t.keys()))
      w = Load_Whiskers( prefix + '.whiskers' ) 
      data = get_summary_data( prefix + '.npy', w, t )
      MeasurementsTable( data ).update_velocities().save( prefix + '.measurements' )


#
# Declarations 
#
cWhisk.Whisker_Segments_Measure.restype = POINTER( cMeasurements )
cWhisk.Whisker_Segments_Measure.argtypes = [
  POINTER( cWhisker_Seg ), # array of whisker segments
  c_int,                   # number of whisker segments
  c_int,                   # face x position (px)
  c_int,                   # face y position (px)
  c_char ]                 # face orientation ( one of: 'h','v','x' or 'y' )

# cWhisk.Whisker_Segments_Measure_With_Bar.restype = POINTER( cMeasurements )
# cWhisk.Whisker_Segments_Measure_With_Bar.argtypes = [
#   POINTER( cWhisker_Seg ), # array of whisker segments
#   c_int,                   # number of whisker segments
#   POINTER( cBar ),         # array of bar locations
#   c_int,                   # number of bar positions
#   c_int,                   # face x position (px)
#   c_int,                   # face y position (px)
#   c_char ]                 # face orientation ( one of: 'h','v','x' or 'y' )

cWhisk.face_point_from_hint.restype = None
cWhisk.face_point_from_hint.argtypes = [
  POINTER( cWhisker_Seg ), # array of whisker segments
  c_int,                   # number of whisker segments
  POINTER( c_char ),       # face hint
  POINTER( c_int  ),       # (out) face x position (px)
  POINTER( c_int  ),       # (out) face y position (px)
  POINTER( c_char ) ]      # (out) face orientation ( one of: 'h','v','x' or 'y' )
  

cWhisk.Measurements_Table_From_Doubles.restype = POINTER(cMeasurements)
cWhisk.Measurements_Table_From_Doubles.argtypes = [
  POINTER( c_double ), # data buffer
  c_int,               # number of rows
  c_int ]              # number of columns

cWhisk.Measurements_Table_Copy_Shape_Data.restype = None
cWhisk.Measurements_Table_Copy_Shape_Data.argtypes = [
  POINTER( cMeasurements ), # the table (the source)
  c_int,                    # number of rows
  POINTER( c_double ) ]     # destination

cWhisk.Measurements_Table_Copy_Velocities.restype = None
cWhisk.Measurements_Table_Copy_Velocities.argtypes = [
  POINTER( cMeasurements ), # the table (the source)
  c_int,                    # number of rows
  POINTER( c_double ) ]     # destination

cWhisk.Measurements_Table_From_Filename.restype = POINTER(cMeasurements)
cWhisk.Measurements_Table_From_Filename.argtypes = [
  POINTER( c_char ),
  POINTER( c_char ),
  POINTER( c_int  ) ]

cWhisk.Alloc_Distributions.restype = POINTER(cDistributions)
cWhisk.Alloc_Distributions.argtypes = [
  c_int,  # n_bins
  c_int,  # n_measures
  c_int ] # n_states

cWhisk.Free_Distributions.restype = None
cWhisk.Free_Distributions.argtypes = [ POINTER(cDistributions) ]

cWhisk.Build_Distributions.restype = POINTER( cDistributions )
cWhisk.Build_Distributions.argtype = [
  POINTER( cMeasurements ), # measurements table
  c_int,                    # number of rows
  c_int ]                   # number of bins

cWhisk.Build_Velocity_Distributions.restype = POINTER( cDistributions )
cWhisk.Build_Velocity_Distributions.argtype = [
  POINTER( cMeasurements ), # measurements table
  c_int,                    # number of rows
  c_int ]                   # number of bins

cWhisk.Solve.restype = None
cWhisk.argtypes = [
  POINTER( cMeasurements ), # table
  c_int,                    # number of rows
  c_int ]                   # number of bins

cWhisk.Measurements_Tables_Get_Diff_Frames.restype = POINTER( c_int )
cWhisk.Measurements_Tables_Get_Diff_Frames.argtypes = [
  POINTER( cMeasurements ), #table A
  c_int,                    #number of rows for table A
  POINTER( cMeasurements ), #table B                   
  c_int,                    #number of rows for table B
  POINTER( c_int ) ]        #size of returned static array

cWhisk.Measurements_Table_Estimate_Best_Threshold.restype = c_double
cWhisk.Measurements_Table_Estimate_Best_Threshold.argtypes = [
  POINTER( cMeasurements ), # table
  c_int,                    # n_rows
  c_int,                    # column index of the feature to use
  c_double,                 # low (px)
  c_double,                 # high (px)
  c_int,                    # is_gt
  POINTER(c_int)            # (output) target count
]


# if __name__=='__main__':
#   testcases = [ 
#                 Tests_MeasurementsTable_FromDoubles,
#                 Tests_MeasurementsTable_FromFile ,
#                 Tests_Distributions 
#                 ]
#   suite = reduce( lambda a,b: a if a.addTest(b) else a, 
#                   list(map( unittest.defaultTestLoader.loadTestsFromTestCase, testcases )) 
#                 )
#   suite.addTest( doctest.DocTestSuite() )
#   runner = unittest.TextTestRunner(verbosity=2,descriptions=1).run(suite)
